"""More cascades

Revision ID: 795ebb9262d8
Revises: 2b5bddd207ae
Create Date: 2024-11-07 19:14:25.263371

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '795ebb9262d8'
down_revision = '2b5bddd207ae'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    try:
        with op.batch_alter_table('certificates', schema=None) as batch_op:
            batch_op.drop_constraint('certificate_eud', type_='foreignkey')
    except:
        pass

    with op.batch_alter_table('certificates', schema=None) as batch_op:
        batch_op.create_foreign_key('certificate_eud', 'euds', ['eud_uid'], ['uid'], ondelete='CASCADE')

    try:
        with op.batch_alter_table('chatrooms_uids', schema=None) as batch_op:
            batch_op.drop_constraint('chatroom_eud', type_='foreignkey')
    except:
        pass

    with op.batch_alter_table('chatrooms_uids', schema=None) as batch_op:
        batch_op.create_foreign_key('chatroom_eud', 'euds', ['uid'], ['uid'], ondelete='CASCADE')

    try:
        with op.batch_alter_table('cot', schema=None) as batch_op:
            batch_op.drop_constraint('cot_eud', type_='foreignkey')
    except:
        pass

    with op.batch_alter_table('cot', schema=None) as batch_op:
        batch_op.create_foreign_key('cot_eud', 'euds', ['sender_uid'], ['uid'], ondelete='CASCADE')

    try:
        with op.batch_alter_table('data_packages', schema=None) as batch_op:
            batch_op.drop_constraint('data_packages_eud', type_='foreignkey')
    except:
        pass

    with op.batch_alter_table('data_packages', schema=None) as batch_op:
        batch_op.create_foreign_key('data_packages_eud', 'euds', ['creator_uid'], ['uid'], ondelete='CASCADE')

    try:
        with op.batch_alter_table('geochat', schema=None) as batch_op:
            batch_op.drop_constraint('geochat_eud', type_='foreignkey')
    except:
        pass

    with op.batch_alter_table('geochat', schema=None) as batch_op:
        batch_op.create_foreign_key('geochat_eud', 'euds', ['sender_uid'], ['uid'], ondelete='CASCADE')

    try:
        with op.batch_alter_table('groups_euds', schema=None) as batch_op:
            batch_op.drop_constraint('group_eud', type_='foreignkey')
    except:
        pass

    with op.batch_alter_table('groups_euds', schema=None) as batch_op:
        batch_op.create_foreign_key('group_eud', 'euds', ['eud_uid'], ['uid'], ondelete='CASCADE')

    try:
        with op.batch_alter_table('mission_invitations', schema=None) as batch_op:
            batch_op.drop_constraint('mission_invitation_callsign', type_='foreignkey')
            batch_op.drop_constraint('mission_invitation_eud', type_='foreignkey')
    except:
        pass

    with op.batch_alter_table('mission_invitations', schema=None) as batch_op:
        batch_op.create_foreign_key('mission_invitation_eud', 'euds', ['client_uid'], ['uid'], ondelete='CASCADE')
        batch_op.create_foreign_key('mission_invitation_callsign', 'euds', ['callsign'], ['callsign'], ondelete='CASCADE')

    try:
        with op.batch_alter_table('missions', schema=None) as batch_op:
            batch_op.drop_constraint('mission_creator', type_='foreignkey')
    except:
        pass

    with op.batch_alter_table('missions', schema=None) as batch_op:
        batch_op.create_foreign_key('mission_creator', 'euds', ['creator_uid'], ['uid'], ondelete='CASCADE')

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('missions', schema=None) as batch_op:
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.create_foreign_key('mission_creator', 'euds', ['creator_uid'], ['uid'])

    with op.batch_alter_table('mission_invitations', schema=None) as batch_op:
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.create_foreign_key('mission_invitation_eud', 'euds', ['client_uid'], ['uid'])
        batch_op.create_foreign_key('mission_invitation_callsign', 'euds', ['callsign'], ['callsign'])

    with op.batch_alter_table('groups_euds', schema=None) as batch_op:
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.create_foreign_key('group_eud', 'euds', ['eud_uid'], ['uid'])

    with op.batch_alter_table('geochat', schema=None) as batch_op:
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.create_foreign_key('geochat_eud', 'euds', ['sender_uid'], ['uid'])

    with op.batch_alter_table('data_packages', schema=None) as batch_op:
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.create_foreign_key('data_packages_eud', 'euds', ['creator_uid'], ['uid'])

    with op.batch_alter_table('cot', schema=None) as batch_op:
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.create_foreign_key('cot_eud', 'euds', ['sender_uid'], ['uid'])

    with op.batch_alter_table('chatrooms_uids', schema=None) as batch_op:
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.create_foreign_key('chatroom_eud', 'euds', ['uid'], ['uid'])

    with op.batch_alter_table('certificates', schema=None) as batch_op:
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.create_foreign_key('certificate_eud', 'euds', ['eud_uid'], ['uid'])

    # ### end Alembic commands ###



from datetime import timedelta, timezone
from uuid import UUID

import pika
import sqlalchemy.exc
from flask import Blueprint, request, jsonify
from flask_babel import gettext
from flask_security import auth_required, current_user
from sqlalchemy import insert, update

from werkzeug.datastructures import ImmutableMultiDict

from opentakserver.blueprints.ots_api.api import search, paginate, route_cot
from opentakserver.extensions import db, logger, socketio
from opentakserver.forms.casevac_form import CasEvacForm
from opentakserver.forms.zmist_form import ZmistForm
from opentakserver.functions import *
from opentakserver.models.CasEvac import CasEvac
from opentakserver.models.CoT import CoT
from opentakserver.models.EUD import EUD
from opentakserver.models.Point import Point
from opentakserver.models.ZMIST import ZMIST

casevac_api_blueprint = Blueprint('casevac_api_blueprint', __name__)


@casevac_api_blueprint.route("/api/casevac", methods=['GET'])
@auth_required()
def query_casevac():
    query = db.session.query(CasEvac)

    query = search(query, EUD, 'callsign')
    query = search(query, CasEvac, 'sender_uid')
    query = search(query, CasEvac, 'uid')

    return paginate(query)


@casevac_api_blueprint.route("/api/casevac", methods=['POST'])
@auth_required()
def add_casevac():
    form = CasEvacForm(formdata=ImmutableMultiDict(request.json))
    if not form.validate():
        return jsonify({'success': False, 'errors': form.errors}), 400

    zmist = None
    if 'zmist' in request.json.keys():
        logger.warning(request.json['zmist'])
        zmist_form = ZmistForm(formdata=ImmutableMultiDict(request.json['zmist']))
        if not zmist_form.validate():
            return jsonify({'success': False, 'errors': form.errors}), 400
        zmist = ZMIST()
        zmist.from_wtform(zmist_form)

    casevac = CasEvac()
    casevac.from_wtforms(form)
    casevac.sender_uid = app.config.get("OTS_NODE_ID")
    if zmist:
        casevac.zmist = zmist

    point = Point()
    point.from_wtform(form)
    point.device_uid = app.config.get("OTS_NODE_ID")

    casevac.point = point

    cot = CoT()
    cot.how = "h-g-i-g-o"
    cot.type = "b-r-f-h-c"
    cot.sender_uid = app.config.get("OTS_NODE_ID")
    cot.sender_callsign = app.config.get("OTS_NODE_ID")
    cot.timestamp = point.timestamp
    cot.start = point.timestamp
    cot.stale = point.timestamp + timedelta(days=365)
    cot.xml = tostring(casevac.to_cot()).decode('utf-8')

    cot_result = db.session.execute(insert(CoT).values(**cot.serialize()))
    db.session.commit()

    cot_pk = cot_result.inserted_primary_key[0]
    point.cot_id = cot_pk
    casevac.cot_id = cot_pk

    point_result = db.session.execute(insert(Point).values(**point.serialize()))
    db.session.commit()
    point_pk = point_result.inserted_primary_key[0]

    casevac.point_id = point_pk

    try:
        db.session.add(casevac)

        if zmist:
            zmist.casevac_uid = casevac.uid
            db.session.add(zmist)
        db.session.commit()
        logger.debug(f"Saved new CasEvac {casevac.uid}")

    except sqlalchemy.exc.IntegrityError:
        db.session.rollback()
        db.session.execute(update(CasEvac).where(CasEvac.uid == casevac.uid).values(
            point_id=point_pk,
            cot_id=casevac.cot_id,
            **casevac.serialize()
        ))
        db.session.commit()
        logger.debug(f"Updated CasEvac {casevac.uid}")

    rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
    rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
    rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
    channel = rabbit_connection.channel()
    channel.basic_publish(exchange='cot', routing_key='', body=json.dumps({'cot': cot.xml, 'uid': app.config['OTS_NODE_ID']}),
                          properties=pika.BasicProperties(expiration=app.config.get("OTS_RABBITMQ_TTL")))
    route_cot(cot.xml, current_user)
    channel.close()
    rabbit_connection.close()

    casevac.zmist = zmist
    socketio.emit('casevac', casevac.to_json(), namespace="/socket.io")

    return jsonify({'success': True}), 200


@casevac_api_blueprint.route("/api/casevac", methods=['DELETE'])
@auth_required()
def delete_casevac():
    uid = request.args.get('uid')
    if not uid:
        return jsonify({'success': False, 'error': 'Please specify a UID'}), 400

    try:
        UUID(uid, version=4)
    except ValueError:
        return jsonify({'success': False, 'error': f"Invalid UID: {uid}"}), 400

    query = db.session.query(CasEvac)
    query = search(query, CasEvac, 'uid')
    casevac = db.session.execute(query).first()
    if not casevac:
        return jsonify({'success': False, 'error': gettext(u'Unknown UID: %(uid)s', uid=uid)}), 404

    casevac = casevac[0]

    now = datetime.now(timezone.utc)
    event = Element('event', {'how': 'h-g-i-g-o', 'type': 't-x-d-d', 'version': '2.0',
                              'uid': casevac.uid, 'start': iso8601_string_from_datetime(now),
                              'time': iso8601_string_from_datetime(now),
                              'stale': iso8601_string_from_datetime(now)})
    SubElement(event, 'point', {'ce': '9999999', 'le': '9999999', 'hae': '0', 'lat': '0',
                                'lon': '0'})
    detail = SubElement(event, 'detail')
    SubElement(detail, 'link', {'relation': 'p-p', 'uid': casevac.uid, 'type': casevac.cot.type})
    SubElement(detail, '_flow-tags_',
               {'TAK-Server-f1a8159ef7804f7a8a32d8efc4b773d0': iso8601_string_from_datetime(now)})

    rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
    rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
    rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
    channel = rabbit_connection.channel()
    channel.basic_publish(exchange='cot', routing_key='', body=json.dumps(
        {'cot': tostring(event).decode('utf-8'), 'uid': app.config['OTS_NODE_ID']}),
                          properties=pika.BasicProperties(expiration=app.config.get("OTS_RABBITMQ_TTL")))
    route_cot(tostring(event).decode('utf-8'), current_user)
    channel.close()
    rabbit_connection.close()

    db.session.delete(casevac)
    db.session.commit()

    return jsonify({'success': True})



import hashlib
import io
import os
import traceback
import uuid
from datetime import datetime, timezone
from urllib.parse import urlparse
import zipfile

import bleach
import sqlalchemy
from flask import Blueprint, request, jsonify, current_app as app, send_from_directory
from flask_babel import gettext
from flask_login import current_user
from werkzeug.utils import secure_filename
from werkzeug.datastructures.file_storage import FileStorage

from xml.etree.ElementTree import Element, tostring, SubElement

from opentakserver.extensions import logger, db
from opentakserver.functions import iso8601_string_from_datetime, format_bytes
from opentakserver.models.DataPackage import DataPackage
from opentakserver.models.MissionContent import MissionContent

data_package_marti_api = Blueprint('data_package_marti_api', __name__)


def save_data_package_to_db(filename: str = None, sha256_hash: str = None, mimetype: str = "application/zip",
                            file_size: int = 0):
    try:
        data_package = DataPackage()
        data_package.filename = filename
        data_package.hash = sha256_hash
        data_package.creator_uid = request.args.get('creatorUid') if request.args.get('creatorUid') else None
        data_package.submission_user = current_user.id if current_user.is_authenticated else None
        data_package.submission_time = datetime.now(timezone.utc)
        data_package.mime_type = mimetype
        data_package.size = file_size
        db.session.add(data_package)
        db.session.commit()
    except sqlalchemy.exc.IntegrityError as e:
        db.session.rollback()
        logger.error("Failed to save data package: {}".format(e))
        logger.debug(traceback.format_exc())
        return jsonify({'success': False, 'error': gettext(u'This data package has already been uploaded')}), 400


def create_data_package_zip(file: FileStorage | str) -> str:
    if isinstance(file, str):
        logger.info(secure_filename(os.path.basename(file)))
        filename, extension = os.path.splitext(secure_filename(os.path.basename(file)))
        logger.info(f"Got filename {filename}")
    else:
        filename, extension = os.path.splitext(secure_filename(file.filename))

    zipf = zipfile.ZipFile(os.path.join(app.config.get("UPLOAD_FOLDER"), f"{filename}.zip"), "a", zipfile.ZIP_DEFLATED,
                           False)

    # Use the md5 of the uploaded file as its folder name in the data package zip
    md5 = hashlib.md5()
    try:
        md5.update(file.stream.read())
        file.stream.seek(0)
    except AttributeError:
        with open(file, "r") as f:
            md5.update(f.read().encode())

    md5_hash = md5.hexdigest()

    if isinstance(file, str):
        with open(file, "r") as f:
            zipf.writestr(f"{md5_hash}/{secure_filename(filename + extension)}", f.read())
    else:
        zipf.writestr(f"{md5_hash}/{secure_filename(filename + extension)}", file.stream.read())

    # MANIFEST file
    manifest = Element("MissionPackageManifest", {"version": "2"})
    config = SubElement(manifest, "Configuration")
    SubElement(config, "Parameter", {"name": "uid", "value": str(uuid.uuid4())})
    SubElement(config, "Parameter", {"name": "name", "value": secure_filename(filename + extension)})

    contents = SubElement(manifest, "Contents")
    content = SubElement(contents, "Content",
                         {"ignore": "false", "zipEntry": f"{md5_hash}/{secure_filename(filename + extension)}"})

    extension = extension.lower().replace(".", "")
    if extension == 'kml' or extension == 'kmz':
        SubElement(content, 'Parameter', {'name': 'name', 'value': f"{secure_filename(filename + extension)}"})
        SubElement(content, 'Parameter', {'name': 'contentType', 'value': 'KML'})
        SubElement(content, 'Parameter', {'name': 'visible', 'value': 'true'})

    zipf.writestr("MANIFEST/manifest.xml", tostring(manifest))
    zipf.close()

    # Get the sha256 hash of the data package zip for its file name on disk and for the data_packages table
    zip_file = open(os.path.join(app.config.get("UPLOAD_FOLDER"), f"{filename}.zip"), 'rb')
    zip_file_bytes = zip_file.read()
    zip_file.close()

    sha256 = hashlib.sha256()
    sha256.update(zip_file_bytes)
    data_package_hash = sha256.hexdigest()

    os.rename(os.path.join(app.config.get("UPLOAD_FOLDER"), f"{filename}.zip"),
              os.path.join(app.config.get("UPLOAD_FOLDER"), f"{data_package_hash}.zip"))
    save_data_package_to_db(f"{filename}.zip", data_package_hash, "application/zip", len(zip_file_bytes))

    return data_package_hash


@data_package_marti_api.route('/Marti/sync/missionupload', methods=['POST'])
def data_package_share():
    if not len(request.files):
        return jsonify({'success': False, 'error': gettext(u'A file is required')}), 400
    for file in request.files:
        file = request.files[file]

        name, extension = os.path.splitext(file.filename)
        extension = extension.replace(".", "")

        # ATAK sends data packages as zips with a file name but no extension
        if not extension and 'zip' in file.mimetype:
            extension = 'zip'

        if extension.lower() not in app.config.get("ALLOWED_EXTENSIONS"):
            logger.info(f"file is {file.filename}, extension is {extension}, content-type {file.mimetype}")
            return jsonify({'success': False, 'error': gettext(u'Invalid file extension: %(extension)s', extension=extension)}), 400

        if extension != 'zip':
            file_hash = create_data_package_zip(file)

        else:
            file_hash = request.args.get('hash')
            if not file_hash:
                sha256 = hashlib.sha256()
                sha256.update(file.stream.read())
                file.stream.seek(0)
                file_hash = sha256.hexdigest()
                logger.debug("got sha256 {}".format(file_hash))

            logger.debug("Got file: {} - {}".format(file.filename, file_hash))

            file.save(os.path.join(app.config.get("UPLOAD_FOLDER"), f'{file_hash}.zip'))

            filename, extension = os.path.splitext(secure_filename(file.filename))
            file_size = os.path.getsize(os.path.join(app.config.get("UPLOAD_FOLDER"), f'{file_hash}.zip'))
            save_data_package_to_db(f'{filename}.zip', file_hash, file.content_type, file_size)

        url = urlparse(request.url_root)
        return 'https://{}:{}/Marti/api/sync/metadata/{}/tool'.format(url.hostname,
                                                                      app.config.get("OTS_MARTI_HTTPS_PORT"),
                                                                      file_hash), 200


@data_package_marti_api.route('/Marti/api/sync/metadata/<file_hash>/tool', methods=['GET', 'PUT'])
def data_package_metadata(file_hash):
    if request.method == 'PUT':
        try:
            data_package = db.session.execute(db.select(DataPackage).filter_by(hash=file_hash)).first()
            if data_package:
                data_package = data_package[0]
                data_package.keywords = bleach.clean(request.data.decode("utf-8"))
                db.session.add(data_package)
                db.session.commit()
                return '', 200
            else:
                return '', 404
        except BaseException as e:
            logger.error("Data package PUT failed: {}".format(e))
            logger.error(traceback.format_exc())
            return {'error': str(e)}, 500
    elif request.method == 'GET':
        data_package = db.session.execute(db.select(DataPackage).filter_by(hash=file_hash)).scalar_one()
        return send_from_directory(app.config.get("UPLOAD_FOLDER"), data_package.hash + ".zip",
                                   download_name=data_package.filename)


@data_package_marti_api.route('/Marti/api/sync/search', methods=['GET'])
def get_data_package():
    data_package_hash = request.args.get("hash")

    # TODO: Support keywords
    keyword = request.args.get("keyword")

    query = db.select(DataPackage)
    if data_package_hash:
        query = query.where(DataPackage.hash == data_package_hash)

    data_packages = db.session.execute(query).scalars()
    return_value = {"version": "3", "type": "gov.tak.api.comms.takserver.mission.data.Resource", "data": [],
                    "messages": [], "nodeId": app.config.get("OTS_NODE_ID")}

    for dp in data_packages:
        submission_user = "anonymous"
        if dp.user:
            submission_user = dp.user.username
        keywords = []
        if dp.keywords:
            keywords = dp.keywords.split(',')

        try:
            expiration = None
            if dp.expiration:
                expiration = int(dp.expiration)
            return_value["data"].append({"filename": dp.filename, "keywords": keywords, "mimeType": dp.mime_type,
                                         "name": dp.filename,
                                         "submissionTime": iso8601_string_from_datetime(dp.submission_time),
                                         "submitter": submission_user, "uid": dp.hash, "creatorUid": dp.creator_uid,
                                         "hash": dp.hash,
                                         "size": dp.size, "tool": dp.tool, "groups": [], "expiration": expiration,
                                         "latitude": 0, "longitude": 0, "altitude": 0})
        except BaseException as e:
            logger.error(f"Data package GET failed: {e}")
            logger.debug(traceback.format_exc())

    return jsonify(return_value)


@data_package_marti_api.route('/Marti/sync/search', methods=['GET'])
def data_package_search():
    data_packages = db.session.execute(db.select(DataPackage)).scalars()
    res = {'resultCount': 0, 'results': []}
    for dp in data_packages:
        submission_user = "anonymous"
        if dp.user:
            submission_user = dp.user.username
        res['results'].append(
            {'UID': dp.hash, 'Name': dp.filename, 'Hash': dp.hash, 'CreatorUid': dp.creator_uid,
             "SubmissionDateTime": dp.submission_time.strftime('%Y-%m-%dT%H:%M:%S.000Z'), "EXPIRATION": "-1",
             "Keywords": ["missionpackage"],
             "MIMEType": dp.mime_type, "Size": "{}".format(dp.size), "SubmissionUser": submission_user,
             "PrimaryKey": "{}".format(dp.id),
             "Tool": dp.tool if dp.tool else "public"
             })
        res['resultCount'] += 1

    return jsonify(res)


@data_package_marti_api.route('/Marti/sync/content', methods=['GET', 'HEAD'])
def download_data_package():
    file_hash = request.args.get('hash')
    file = db.session.execute(db.select(DataPackage).filter_by(hash=file_hash)).first()
    if not file:
        file = db.session.execute(db.select(MissionContent).filter_by(hash=file_hash)).first()
    if not file and request.method == 'HEAD':
        return '', 404
    elif not file:
        return jsonify({'success': False, 'error': f'No data package found with hash {file_hash}'}), 404
    elif file and request.method == 'HEAD':
        return '', 200

    filename, extension = os.path.splitext(secure_filename(file[0].filename))
    if os.path.exists(os.path.join(app.config.get("UPLOAD_FOLDER"), f"{file_hash}{extension}")):
        return send_from_directory(app.config.get("UPLOAD_FOLDER"), f"{file_hash}{extension}",
                                   download_name=file[0].filename)
    elif os.path.exists(os.path.join(app.config.get("OTS_DATA_FOLDER"), "missions", file[0].filename)):
        return send_from_directory(os.path.join(app.config.get("OTS_DATA_FOLDER"), "missions"), file[0].filename)
    else:
        return jsonify({'success': False, 'error': gettext(u'File not found: %(filename)s}', filename=file[0].filename)}), 404


@data_package_marti_api.route('/Marti/sync/missionquery')
def data_package_query():
    try:
        data_package = db.session.execute(db.select(DataPackage).filter_by(hash=request.args.get('hash'))).scalar_one()
        if data_package:

            url = urlparse(request.url_root)
            return 'https://{}:{}/Marti/api/sync/metadata/{}/tool'.format(url.hostname,
                                                                          app.config.get("OTS_MARTI_HTTPS_PORT"),
                                                                          request.args.get('hash')), 200
        else:
            return jsonify({'success': False, 'error': 'File not found'}), 404
    except sqlalchemy.exc.NoResultFound as e:
        return jsonify({'success': False, 'error': 'File not found'}), 404


@data_package_marti_api.route('/Marti/api/files/metadata')
def get_metadata():
    mission_package = request.args.get('missionPackage')
    page = request.args.get('page')
    limit = request.args.get('limit')
    mission = request.args.get('mission')
    name = request.args.get('name')
    sort = request.args.get('sort')
    setting = request.args.get('setting')

    query = db.session.query(DataPackage)
    if name:
        query = query.filter_by(filename=bleach.clean(name))

    data_packages = db.session.execute(query).scalars()

    response = {"version": "3", "type": "gov.tak.api.comms.takserver.mission.data.Resource", "data": [],
                "nodeId": app.config.get("OTS_NODE_ID")}
    for dp in data_packages:
        response['data'].append({
            "Name": dp.filename,
            "User": dp.user.username if dp.user else None,
            "Creator": dp.eud.callsign if dp.eud else None,
            "Size": format_bytes(dp.size),
            "Time": iso8601_string_from_datetime(dp.submission_time),
            "MimeType": dp.mime_type,
            "Keywords": dp.keywords,
            "Expiration": dp.expiration,
            "Hash": dp.hash
        })

    return jsonify(response)



"""Added groups_missions table

Revision ID: 2b1863a985a6
Revises: eaf2af6f6b22
Create Date: 2025-11-09 18:33:51.885010

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '2b1863a985a6'
down_revision = 'eaf2af6f6b22'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('groups_missions',
    sa.Column('mission_name', sa.String(length=255), nullable=False),
    sa.Column('group_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['group_id'], ['groups.id'], ),
    sa.ForeignKeyConstraint(['mission_name'], ['missions.name'], ),
    sa.PrimaryKeyConstraint('mission_name', 'group_id')
    )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    op.drop_table('groups_missions')
    # ### end Alembic commands ###



from __future__ import annotations

import os
import subprocess
import sys
import traceback
from typing import TYPE_CHECKING

import selectors
import sqlalchemy.exc
from flask import Flask, request, current_app as app
from sqlalchemy import update

from opentakserver.blueprints.ots_socketio import administrator_only
from opentakserver.extensions import logger, db, socketio
from opentakserver.models.Plugins import Plugins
from opentakserver.plugins.Plugin import Plugin
from poetry.utils._compat import metadata
from werkzeug.utils import secure_filename


if TYPE_CHECKING:
    from typing import Any


class PluginManager:
    """
    This class registers and activates plugins.
    """

    def __init__(self, group: str, app: Flask) -> None:
        self._group = group
        self.plugins: dict[str, Plugin] = {}
        self._app = app

    def load_plugins(self) -> None:
        plugin_entrypoints = self.get_plugin_entry_points()
        for ep in plugin_entrypoints:
            self._load_plugin_entry_point(ep)

    def get_plugin_entry_points(self) -> list[metadata.EntryPoint]:
        return list(metadata.entry_points(group=self._group))

    def activate(self, *args: Any, **kwargs: Any) -> None:
        for name, plugin in self.plugins.items():
            try:
                try:
                    with self._app.app_context():
                        plugin_metadata = plugin.load_metadata()

                        # Add the plugin to the DB or update its version number
                        plugin_row = db.session.execute(db.session.query(Plugins).filter_by(name=plugin.name.lower())).first()
                        if plugin_row:
                            plugin_row = plugin_row[0]
                            plugin_row.version = plugin_metadata.get("version")
                        else:
                            plugin_row = Plugins()
                            plugin_row.name = plugin.name.lower()
                            plugin_row.distro = plugin_metadata.get("distro")
                            plugin_row.author = plugin_metadata.get("author")
                            plugin_row.version = plugin_metadata.get("version")
                            plugin_row.enabled = True

                        db.session.add(plugin_row)
                        db.session.commit()
                except sqlalchemy.exc.IntegrityError as e:
                    logger.debug(f"{plugin_row.name} is already in the DB")
                except BaseException as e:
                    logger.debug(f"{plugin.name} already in the database")

                plugin.activate(*args, **kwargs, enabled=self.check_if_plugin_enabled(plugin.name))
                if plugin.blueprint:
                    self._app.register_blueprint(plugin.blueprint)

            except BaseException as e:
                logger.error(f"Failed to load plugin: {e}")
                logger.error(traceback.format_exc())

    def stop_plugins(self):
        for name, plugin in self.plugins.items():
            plugin.stop()

    def disable_plugin(self, plugin_distro: str):
        plugin = self.plugins[plugin_distro]
        plugin.stop()

        db.session.execute(update(Plugins).where(Plugins.name == plugin.name.lower()).values(enabled=False))
        db.session.commit()

        logger.info(f"{plugin_distro} disabled")

    def enable_plugin(self, plugin_distro: str):
        plugin = self.plugins[plugin_distro]
        plugin.activate(self._app, True)

        db.session.execute(update(Plugins).where(Plugins.name == plugin.name.lower()).values(enabled=True))
        db.session.commit()

        logger.info(f"{plugin_distro} enabled")

    def _add_plugin(self, plugin: Plugin) -> None:
        if not isinstance(plugin, Plugin):
            raise ValueError(
                "The OTS plugin must be an instance of Plugin"
            )

        logger.info(f"Adding {plugin.name}")
        plugin.load_metadata()
        self.plugins[plugin.distro.lower()] = plugin

    def _load_plugin_entry_point(self, ep: metadata.EntryPoint) -> None:
        plugin = ep.load()

        if not issubclass(plugin, Plugin):
            raise ValueError(
                "The OTS plugin must be an instance of Plugin"
            )

        try:
            self._add_plugin(plugin())
        except BaseException as e:
            logger.error(f"Failed to load plugin: {e}")
            logger.error(traceback.format_exc())

    def check_if_plugin_enabled(self, plugin_distro: str) -> bool:
        with self._app.app_context():
            plugin = db.session.execute(db.session.query(Plugins).filter_by(name=plugin_distro.lower())).first()
            if plugin:
                return plugin[0].enabled
            else:
                # First time this plugin has been loaded, enable it by default
                return True

    @staticmethod
    @socketio.on('plugin_package_manager', namespace="/socket.io")
    @administrator_only
    def install_plugin(json: dict):
        plugin_valid = json.get('action') == 'install_local'
        file_name = None

        if json.get('action') != "install_local":
            for plugin_prefix in app.config.get("OTS_PLUGIN_PREFIXES"):
                if json.get('plugin_distro').startswith(plugin_prefix):
                    plugin_valid = True
                    break

        if not plugin_valid:
            logger.error(f"Invalid Plugin: {json.get('plugin_distro')}")
            socketio.emit('plugin_package_manager', {"success": False, "message": f"Invalid Plugin: {json.get('plugin_distro')}"}, to=request.sid, namespace="/socket.io")
            return

        if json.get('action') == 'delete':
            command = f"{sys.executable} -m pip uninstall --yes {json.get('plugin_distro')}"
            try:
                logger.info(f"Disabling plugin {json.get('plugin_distro')}")
                app.plugin_manager.disable_plugin(json.get('plugin_distro').lower())
            except BaseException as e:
                logger.error(f"Failed to disable plugin: {e}")
                logger.debug(traceback.format_exc())

        elif json.get('action') == 'install':
            command = f"{sys.executable} -m pip --no-input install {json.get('plugin_distro')} -i {app.config.get('OTS_PLUGIN_REPO')}"

        elif json.get('action') == 'install_local':
            logger.info(f"{json.get('action')} {json.get('plugin_file_name')}")
            file_name = secure_filename(json.get('plugin_file_name'))

            if not os.path.exists(os.path.join(app.config.get('UPLOAD_FOLDER'), file_name)):
                logger.error(f"Plugin file not found: {file_name}")
                socketio.emit('plugin_package_manager', {"success": False, "message": f"Plugin file not found: {file_name}"}, namespace="/socket.io", to=request.sid, ignore_queue=True)
                return

            command = f"{sys.executable} -m pip --no-input install {os.path.join(app.config.get('UPLOAD_FOLDER'), file_name)}"

        else:
            logger.error(f"Invalid action: {json.get('action')}")
            socketio.emit('plugin_package_manager', {"success": False, "message": f"Invalid action: {json.get('action')}"}, namespace="/socket.io", to=request.sid, ignore_queue=True)
            return

        socketio.emit('plugin_package_manager', {"message": f"$ {command}\n"}, namespace="/socket.io", to=request.sid, ignore_queue=True)
        return_code = None

        try:
            output = subprocess.Popen(command.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE)
            sel = selectors.DefaultSelector()
            sel.register(output.stdout, selectors.EVENT_READ)
            sel.register(output.stderr, selectors.EVENT_READ)
        except BaseException as e:
            logger.error(f"Failed to run command {command}: {e}")
            logger.error(traceback.format_exc())
            socketio.emit("plugin_package_manager", {"success": False, "message": f"Command failed: {e}"}, namespace="/socket.io", to=request.sid, ignore_queue=True)
            return

        while return_code is None:
            for key, _ in sel.select():
                if key.fileobj != output.stdin:
                    data = key.fileobj.read1().decode()

                    if data:
                        socketio.emit('plugin_package_manager', {"message": data}, to=request.sid, namespace="/socket.io", ignore_queue=True)

                return_code = output.poll()
        if return_code == 0:
            socketio.emit('plugin_package_manager', {"success": True, "message": "Command completed successfully"}, to=request.sid, namespace="/socket.io", ignore_queue=True)

            if json.get('action') == 'install':
                entry_points = app.plugin_manager.get_plugin_entry_points()
                for entry_point in entry_points:
                    plugin = entry_point.load()
                    if plugin().distro.lower() == json.get("plugin_distro").lower():
                        app.plugin_manager._add_plugin(plugin())
                        plugin_row = Plugins()
                        plugin_row.name = plugin().name.lower()
                        plugin_row.distro = plugin().distro
                        plugin_row.author = plugin().metadata['author']
                        plugin_row.version = plugin().metadata['version']
                        plugin_row.enabled = True
                        db.session.add(plugin_row)
                        db.session.commit()

                        if json.get('plugin_distro').lower() not in app.plugin_manager.plugins:
                            app.plugin_manager.plugins[json.get('plugin_distro').lower()] = plugin
                        break

                app.plugin_manager.enable_plugin(json.get("plugin_distro").lower())

            elif json.get('action') == 'delete':
                del app.plugin_manager.plugins[json.get('plugin_distro').lower()]
                plugin = db.session.execute(db.session.query(Plugins).filter_by(distro=json.get('plugin_distro').lower())).first()
                if plugin:
                    db.session.delete(plugin[0])
                    db.session.commit()
            else:
                app.plugin_manager.load_plugins()
        else:
            logger.error(f"Action {json.get('action')} failed with exit code: {return_code}")
            socketio.emit('plugin_package_manager', {"success": False, "message": f"Command failed with return code: {return_code}"}, to=request.sid, namespace="/socket.io", ignore_queue=True)

        if json.get('action') == 'install_local' and file_name is not None:
            os.remove(os.path.join(app.config.get('UPLOAD_FOLDER'), file_name))



from flask_wtf import FlaskForm
from wtforms import StringField, IntegerField, BooleanField, FloatField, DateTimeField
from wtforms.validators import DataRequired, UUID, NumberRange, Optional


class PointForm(FlaskForm):
    latitude = FloatField(validators=[DataRequired(), NumberRange(min=-90, max=90)])
    longitude = FloatField(validators=[DataRequired(), NumberRange(min=-180, max=180)])
    ce = FloatField(default=9999999.0)
    hae = FloatField(default=9999999.0)
    le = FloatField(default=9999999.0)
    course = FloatField(default=0.0, validators=[NumberRange(min=0)])
    speed = FloatField(default=0.0, validators=[NumberRange(min=0)])
    location_source = StringField()
    battery = FloatField(validators=[NumberRange(min=0, max=100), Optional()], default=None)
    timestamp = DateTimeField(format="%Y-%m-%dT%H:%M:%S.%fZ", validators=[DataRequired()])
    azimuth = FloatField(validators=[NumberRange(min=0, max=360), Optional()], default=None)
    fov = FloatField(validators=[NumberRange(min=0, max=360), Optional()], default=None)



"""Added columns to mission_invtations table

Revision ID: f6dfc571d31c
Revises: 14e268184cc5
Create Date: 2024-10-17 18:53:11.141218

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'f6dfc571d31c'
down_revision = '14e268184cc5'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table('mission_invitations', schema=None) as batch_op:
        batch_op.add_column(sa.Column('creator_uid', sa.String(length=255), nullable=True))
        batch_op.add_column(sa.Column('role', sa.String(length=255), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('mission_invitations', schema=None) as batch_op:
        batch_op.drop_column('role')
        batch_op.drop_column('creator_uid')

    # ### end Alembic commands ###



"""Made cot.sender_callsign nullable

Revision ID: 85dbe0a7dcc1
Revises: 6665d030f187
Create Date: 2025-01-30 15:19:53.796637

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '85dbe0a7dcc1'
down_revision = '6665d030f187'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('cot', schema=None) as batch_op:
        batch_op.alter_column('sender_callsign',
               existing_type=sa.VARCHAR(length=255),
               nullable=True)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('cot', schema=None) as batch_op:
        batch_op.alter_column('sender_callsign',
               existing_type=sa.VARCHAR(length=255),
               nullable=False)

    # ### end Alembic commands ###



import secrets
import string
import random

import pyotp
from pathlib import Path
import os


class DefaultConfig:
    SECRET_KEY = os.getenv("SECRET_KEY", secrets.token_hex())
    DEBUG = os.getenv("DEBUG", "False").lower() in ["true", "1", "yes"]

    OTS_LANGUAGES = {'US': {'name': 'English', 'language_code': 'en'},
                     'DE': {'name': 'Deutsch', 'language_code': 'de'},
                     'FR': {'name': 'Français', 'language_code': 'fr'},
                     'PT': {'name': 'Português', 'language_code': 'pt'},
                     'ES': {'name': 'Español', 'language_code': 'es'},
                     'DK': {'name': 'dansk', 'language_code': 'da'},
                     'UA': {'name': 'українська', 'language_code': 'uk'},
                     'KR': {'name': '한국어', 'language_code': 'ko'},
                     'PL': {'name': 'Polski', 'language_code': 'pl'},
                     'BR': {'name': 'Português', 'language_code': 'pt_BR'},
                     }

    OTS_DATA_FOLDER = os.getenv("OTS_DATA_FOLDER", os.path.join(Path.home(), "ots"))
    OTS_LISTENER_ADDRESS = os.getenv("OTS_LISTENER_ADDRESS", "127.0.0.1")
    OTS_LISTENER_PORT = int(os.getenv("OTS_LISTENER_PORT", 8081))
    OTS_MARTI_HTTP_PORT = int(os.getenv("OTS_MARTI_HTTP_PORT", 8080))
    OTS_MARTI_HTTPS_PORT = int(os.getenv("OTS_MARTI_HTTPS_PORT", 8443))
    OTS_ENABLE_TCP_STREAMING_PORT = os.getenv("OTS_ENABLE_TCP_STREAMING_PORT", "True").lower() in ["true", "1", "yes"]
    OTS_TCP_STREAMING_PORT = int(os.getenv("OTS_TCP_STREAMING_PORT", 8088))
    OTS_SSL_STREAMING_PORT = int(os.getenv("OTS_SSL_STREAMING_PORT", 8089))
    OTS_BACKUP_COUNT = int(os.getenv("OTS_BACKUP_COUNT", 7))
    OTS_ENABLE_CHANNELS = os.getenv("OTS_ENABLE_CHANNELS", "True").lower() in ["true", "1", "yes"]

    # RabbitMQ Settings
    OTS_RABBITMQ_SERVER_ADDRESS = os.getenv("OTS_RABBITMQ_SERVER_ADDRESS", "127.0.0.1")
    OTS_RABBITMQ_USERNAME = os.getenv("OTS_RABBITMQ_USERNAME", "guest")
    OTS_RABBITMQ_PASSWORD = os.getenv("OTS_RABBITMQ_PASSWORD", "guest")
    # Messages queued in RabbitMQ will auto-delete after 1 day if not consumed https://www.rabbitmq.com/docs/ttl
    # Set to '0' to disable auto-deletion
    OTS_RABBITMQ_TTL = '86400000'
    # How many CoT messages that cot_parser processes should prefetch. https://www.rabbitmq.com/docs/consumer-prefetch
    OTS_RABBITMQ_PREFETCH = 2

    # TAK.gov account link settings
    OTS_TAK_GOV_LINKED = False
    OTS_TAK_GOV_ACCESS_TOKEN = ""
    OTS_TAK_GOV_REFRESH_TOKEN = ""

    OTS_MEDIAMTX_ENABLE = os.getenv("OTS_MEDIAMTX_ENABLE", "True").lower() in ["true", "1", "yes"]
    OTS_MEDIAMTX_API_ADDRESS = os.getenv("OTS_MEDIAMTX_API_ADDRESS", "http://localhost:9997")
    OTS_MEDIAMTX_TOKEN = os.getenv("OTS_MEDIAMTX_TOKEN", secrets.token_urlsafe(30 * 3 // 4))
    OTS_SSL_VERIFICATION_MODE = int(os.getenv("OTS_SSL_VERIFICATION_MODE", 2))
    OTS_SSL_CERT_HEADER = os.getenv("OTS_SSL_CERT_HEADER", "X-Ssl-Cert")
    OTS_NODE_ID = os.getenv("OTS_NODE_ID", ''.join(random.choices(string.ascii_lowercase + string.digits, k=32)))

    # Certificate Authority Settings
    OTS_CA_NAME = os.getenv("OTS_CA_NAME", "OpenTAKServer-CA")
    OTS_CA_FOLDER = os.getenv("OTS_CA_FOLDER", os.path.join(OTS_DATA_FOLDER, "ca"))
    OTS_CA_PASSWORD = os.getenv("OTS_CA_PASSWORD", "atakatak")
    OTS_CA_EXPIRATION_TIME = int(os.getenv("OTS_CA_EXPIRATION_TIME", 3650))
    OTS_CA_COUNTRY = os.getenv("OTS_CA_COUNTRY", "WW")
    OTS_CA_STATE = os.getenv("OTS_CA_STATE", "XX")
    OTS_CA_CITY = os.getenv("OTS_CA_CITY", "YY")
    OTS_CA_ORGANIZATION = os.getenv("OTS_CA_ORGANIZATION", "ZZ")
    OTS_CA_ORGANIZATIONAL_UNIT = os.getenv("OTS_CA_ORGANIZATIONAL_UNIT", "OpenTAKServer")
    OTS_CA_SUBJECT = os.getenv("OTS_CA_SUBJECT",
                               f"/C={OTS_CA_COUNTRY}/ST={OTS_CA_STATE}/L={OTS_CA_CITY}/O={OTS_CA_ORGANIZATION}/OU={OTS_CA_ORGANIZATIONAL_UNIT}")

    OTS_COT_PARSER_PROCESSES = int(os.getenv("OTS_COT_PARSER_PROCESSES", 1))

    OTS_ENABLE_LDAP = False
    # LDAP users in this group will be considered OTS administrators
    OTS_LDAP_ADMIN_GROUP = "ots_admin"

    # Attributes to control a user's team color, role, and callsign. The default values match takserver's attributes
    OTS_LDAP_COLOR_ATTRIBUTE = "colorAttribute"
    OTS_LDAP_ROLE_ATTRIBUTE = "roleAttribute"
    OTS_LDAP_CALLSIGN_ATTRIBUTE = "callsignAttribute"

    # LDAP user attributes with this prefix can be used to control ATAK settings for a specific user
    OTS_LDAP_PREFERENCE_ATTRIBUTE_PREFIX = "ots_"
    OTS_LDAP_GROUP_PREFIX = "ots_"

    # Flask-LDAP3-Login settings
    LDAP_HOST = "127.0.0.1"
    LDAP_BASE_DN = ""
    LDAP_USER_DN = ""
    LDAP_GROUP_DN = ""
    LDAP_BIND_USER_DN = "cn=admin,ou=users=dc=example,dc=com"
    LDAP_BIND_USER_PASSWORD = "password"

    # See https://docs.python.org/3/library/logging.handlers.html#logging.handlers.TimedRotatingFileHandler
    OTS_LOG_ROTATE_WHEN = os.getenv("OTS_LOG_ROTATE_WHEN", "midnight")
    OTS_LOG_ROTATE_INTERVAL = int(os.getenv("OTS_LOG_ROTATE_INTERVAL", 0))

    # ADS-B Settings
    OTS_AIRPLANES_LIVE_LAT = 40.744213
    OTS_AIRPLANES_LIVE_LON = -73.986939
    OTS_AIRPLANES_LIVE_RADIUS = 10

    OTS_ADSB_GROUP = "ADS-B"
    OTS_AIS_GROUP = "AIS"

    OTS_ENABLE_PLUGINS = True
    OTS_PLUGIN_REPO = "https://repo.opentakserver.io/brian/prod/"
    OTS_PLUGIN_PREFIXES = ["ots-", "ots_"]

    # AIS Settings
    OTS_AISHUB_USERNAME = None
    OTS_AISHUB_SOUTH_LAT = None
    OTS_AISHUB_WEST_LON = None
    OTS_AISHUB_NORTH_LAT = None
    OTS_AISHUB_EAST_LON = None
    OTS_AISHUB_MMSI_LIST = ""
    OTS_AISHUB_IMO_LIST = ""

    OTS_PROFILE_MAP_SOURCES = True

    OTS_ENABLE_MUMBLE_AUTHENTICATION = False

    OTS_IP_WHITELIST = ["127.0.0.1"]

    # Meshtastic settings
    OTS_ENABLE_MESHTASTIC = False
    OTS_MESHTASTIC_TOPIC = "opentakserver"
    OTS_MESHTASTIC_PUBLISH_INTERVAL = 30
    OTS_MESHTASTIC_DOWNLINK_CHANNELS = []
    OTS_MESHTASTIC_NODEINFO_INTERVAL = 3
    OTS_MESHTASTIC_GROUP = "Meshtastic"

    # Email settings
    OTS_ENABLE_EMAIL = os.getenv("OTS_ENABLE_EMAIL", "False").lower() in ["true", "1", "yes"]
    MAIL_SERVER = os.getenv("MAIL_SERVER", "smtp.gmail.com")
    MAIL_PORT = int(os.getenv("MAIL_PORT", 587))
    MAIL_USE_SSL = os.getenv("MAIL_USE_SSL", "False").lower() in ["true", "1", "yes"]
    MAIL_USE_TLS = os.getenv("MAIL_USE_TLS", "True").lower() in ["true", "1", "yes"]
    MAIL_USERNAME = os.getenv("MAIL_USERNAME", None)
    MAIL_PASSWORD = os.getenv("MAIL_PASSWORD", None)
    MAIL_DEBUG = False
    MAIL_DEFAULT_SENDER = None
    MAIL_MAX_EMAILS = None
    MAIL_SUPPRESS_SEND = False
    MAIL_ASCII_ATTACHMENTS = False
    OTS_EMAIL_DOMAIN_WHITELIST = []
    OTS_EMAIL_DOMAIN_BLACKLIST = []
    OTS_EMAIL_TLD_WHITELIST = []
    OTS_EMAIL_TLD_BLACKLIST = []

    OTS_DELETE_OLD_DATA_SECONDS = int(os.getenv("OTS_DELETE_OLD_DATA_SECONDS", 0))
    OTS_DELETE_OLD_DATA_MINUTES = int(os.getenv("OTS_DELETE_OLD_DATA_MINUTES", 0))
    OTS_DELETE_OLD_DATA_HOURS = int(os.getenv("OTS_DELETE_OLD_DATA_HOURS", 0))
    OTS_DELETE_OLD_DATA_DAYS = int(os.getenv("OTS_DELETE_OLD_DATA_DAYS", 0))
    OTS_DELETE_OLD_DATA_WEEKS = int(os.getenv("OTS_DELETE_OLD_DATA_WEEKS", 1))

    # flask-sqlalchemy
    SQLALCHEMY_DATABASE_URI = os.getenv("SQLALCHEMY_DATABASE_URI",
                                        f"postgresql+psycopg://ots:POSTGRESQL_PASSWORD@127.0.0.1/ots")
    SQLALCHEMY_ECHO = os.getenv("SQLALCHEMY_ECHO", "False").lower() in ["true", "1", "yes"]
    SQLALCHEMY_ENGINE_OPTIONS = {"pool_pre_ping": True}
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    SQLALCHEMY_RECORD_QUERIES = False

    ALLOWED_EXTENSIONS = os.getenv("ALLOWED_EXTENSIONS",
                                   "zip,xml,txt,pdf,png,jpg,jpeg,gif,kml,kmz,p12,tif,sqlite").split(",")

    UPLOAD_FOLDER = os.getenv("UPLOAD_FOLDER", os.path.join(OTS_DATA_FOLDER, "uploads"))
    if not os.path.exists(UPLOAD_FOLDER):
        os.makedirs(UPLOAD_FOLDER)

    # Flask-Security-Too
    SECURITY_PASSWORD_SALT = os.getenv("SECURITY_PASSWORD_SALT", str(secrets.SystemRandom().getrandbits(128)))
    REMEMBER_COOKIE_SAMESITE = "strict"
    SESSION_COOKIE_SAMESITE = "strict"
    SECURITY_USERNAME_ENABLE = True
    SECURITY_USERNAME_REQUIRED = True
    SECURITY_TRACKABLE = True
    SECURITY_CSRF_COOKIE_NAME = "XSRF-TOKEN"
    WTF_CSRF_TIME_LIMIT = None
    SECURITY_CSRF_IGNORE_UNAUTH_ENDPOINTS = True
    WTF_CSRF_CHECK_DEFAULT = False
    SECURITY_RETURN_GENERIC_RESPONSES = True
    SECURITY_URL_PREFIX = "/api"
    SECURITY_CHANGEABLE = True
    SECURITY_CHANGE_URL = "/password/change"
    SECURITY_RESET_URL = "/password/reset"
    SECURITY_PASSWORD_LENGTH_MIN = 8
    SECURITY_PASSWORD_CONFIRM_REQUIRED = False
    SECURITY_REGISTERABLE = OTS_ENABLE_EMAIL
    SECURITY_CONFIRMABLE = OTS_ENABLE_EMAIL
    SECURITY_RECOVERABLE = OTS_ENABLE_EMAIL
    SECURITY_TWO_FACTOR = True
    SECURITY_TOTP_SECRETS = {1: os.getenv("SECURITY_TOTP_SECRET", pyotp.random_base32())}
    SECURITY_TOTP_ISSUER = os.getenv("SECURITY_TOTP_ISSUER", "OpenTAKServer")
    SECURITY_TWO_FACTOR_ENABLED_METHODS = ["authenticator", "email"]
    SECURITY_TWO_FACTOR_RESCUE_MAIL = MAIL_USERNAME
    SECURITY_TWO_FACTOR_ALWAYS_VALIDATE = False
    SECURITY_CSRF_PROTECT_MECHANISMS = ["session", "basic"]
    SECURITY_LOGIN_WITHOUT_CONFIRMATION = True
    SECURITY_POST_CONFIRM_VIEW = "/login"
    SECURITY_REDIRECT_BEHAVIOR = 'spa'
    SECURITY_RESET_VIEW = '/reset'
    SECURITY_USERNAME_MIN_LENGTH = 1
    SECURITY_MSG_USERNAME_DISALLOWED_CHARACTERS = (
    "Username can contain only letters, numbers, underscores, and periods", "error")

    SCHEDULER_API_ENABLED = False
    JOBS = [
        {
            "id": "get_airplanes_live_data",
            "func": "opentakserver.blueprints.scheduled_jobs:get_airplanes_live_data",
            "trigger": "interval",
            "seconds": 0,
            "minutes": 1,
            "next_run_time": None
        },
        {
            "id": "delete_video_recordings",
            "func": "opentakserver.blueprints.scheduled_jobs:delete_video_recordings",
            "trigger": "interval",
            "seconds": 0,
            "minutes": 1,
            "next_run_time": None
        },
        {
            "id": "purge_data",
            "func": "opentakserver.blueprints.scheduled_jobs:purge_data",
            "trigger": "cron",
            "day": "*",
            "hour": 0,
            "minute": 0,
            "next_run_time": None
        },
        {
            "id": "ais",
            "func": "opentakserver.blueprints.scheduled_jobs:get_aishub_data",
            "trigger": "interval",
            "seconds": 0,
            "minutes": 1,
            "next_run_time": None
        },
        {
            "id": "delete_old_data",
            "func": "opentakserver.blueprints.scheduled_jobs:delete_old_data",
            "trigger": "interval",
            "seconds": 0,
            "minutes": 1,
            "next_run_time": None
        }
    ]



import os
import shutil
import traceback
from datetime import datetime, timezone

from flask import Blueprint, request, jsonify, current_app as app, send_from_directory
from flask_babel import gettext
from flask_security import auth_required
from sqlalchemy import update
from werkzeug.datastructures import ImmutableMultiDict

from opentakserver.blueprints.ots_api.api import search, paginate
from opentakserver.blueprints.marti_api.data_package_marti_api import data_package_share
from opentakserver.extensions import logger, db
from opentakserver.forms.data_package_form import DataPackageUpdateForm
from opentakserver.models.Certificate import Certificate
from opentakserver.models.DataPackage import DataPackage

data_package_api = Blueprint('data_package_api', __name__)


@data_package_api.route('/api/data_packages', methods=['PATCH'])
@auth_required()
def edit_data_package():
    form = DataPackageUpdateForm(formdata=ImmutableMultiDict(request.json))
    if not form.validate():
        return jsonify({'success': False, 'errors': form.errors}), 400

    data_package = db.session.execute(db.session.query(DataPackage).filter_by(hash=form.hash.data)).first()
    if not data_package:
        return jsonify({'success': False, 'error': f"Package with hash {form.hash.data} not found"}), 404

    data_package = data_package[0]

    if data_package.filename.endswith("_CONFIG.zip"):
        return jsonify({'success': False, 'error': gettext(u"Server connection data packages can't be installed on enrollment or connection")}), 400

    if form.install_on_enrollment.data is not None:
        data_package.install_on_enrollment = form.install_on_enrollment.data
    if form.install_on_connection.data is not None:
        data_package.install_on_connection = form.install_on_connection.data

    data_package.submission_time = datetime.now(timezone.utc)

    db.session.execute(update(DataPackage).filter(DataPackage.hash == data_package.hash).values(**data_package.serialize()))
    db.session.commit()

    return jsonify({'success': True})


@data_package_api.route('/api/data_packages', methods=['DELETE'])
@auth_required()
def delete_data_package():
    file_hash = request.args.get('hash')
    if not file_hash:
        return jsonify({'success': False, 'error': 'Please provide a file hash'}), 400

    query = db.session.query(DataPackage)
    query = search(query, DataPackage, 'hash')
    data_package = db.session.execute(query).first()
    if not data_package:
        return jsonify({'success': False, 'error': gettext(u'Invalid/unknown hash')}), 400

    try:
        logger.warning("Deleting data package {} - {}".format(data_package[0].filename, data_package[0].hash))
        db.session.delete(data_package[0])
        db.session.commit()
        os.remove(os.path.join(app.config.get("UPLOAD_FOLDER"), "{}.zip".format(data_package[0].hash)))

        if data_package[0].certificate:
            Certificate.query.filter_by(id=data_package[0].certificate.id).delete()
            db.session.commit()
            shutil.rmtree(
                os.path.join(app.config.get("OTS_CA_FOLDER"), "certs", data_package[0].certificate.common_name),
                ignore_errors=True)
    except BaseException as e:
        logger.error("Failed to delete data package")
        logger.error(traceback.format_exc())
        return jsonify({'success': False, 'error': str(e)}), 500

    return jsonify({'success': True})


@data_package_api.route('/api/data_packages')
@auth_required()
def data_packages():
    query = db.session.query(DataPackage)
    query = search(query, DataPackage, 'filename')
    query = search(query, DataPackage, 'hash')
    query = search(query, DataPackage, 'creator_uid')
    query = search(query, DataPackage, 'keywords')
    query = search(query, DataPackage, 'mime_type')
    query = search(query, DataPackage, 'size')
    query = search(query, DataPackage, 'tool')

    return paginate(query)


@data_package_api.route('/api/data_packages/download')
@auth_required()
def data_package_download():
    if 'hash' not in request.args.keys():
        return jsonify({'success': False, 'error': gettext(u'Please provide a data package hash')}), 400

    file_hash = request.args.get('hash')

    query = db.session.query(DataPackage)
    query = search(query, DataPackage, 'hash')

    data_package = db.session.execute(query).first()

    if not data_package:
        return jsonify({'success': False, 'error': gettext(u"Data package with hash '%(hash)s' not found", hash=file_hash)}), 404

    download_name = data_package[0].filename
    name, extension = os.path.splitext(download_name)

    return send_from_directory(app.config.get("UPLOAD_FOLDER"), f"{file_hash}{extension}", as_attachment=True,
                               download_name=download_name)


@data_package_api.route('/api/data_packages', methods=['POST'])
@auth_required()
def upload_data_package():
    return data_package_share()



"""More data sync stuff

Revision ID: d1f5df78eace
Revises: 4f0173cdb93b
Create Date: 2024-10-15 03:02:43.059600

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd1f5df78eace'
down_revision = '4f0173cdb93b'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('mission_changes', schema=None) as batch_op:
        batch_op.add_column(sa.Column('content_uid', sa.String(length=255), nullable=True))
        batch_op.create_foreign_key("content_resource", 'mission_content', ['content_uid'], ['uid'])

    with op.batch_alter_table('mission_content', schema=None) as batch_op:
        batch_op.add_column(sa.Column('filename', sa.String(length=255), nullable=True))
        batch_op.drop_column('name')

    with op.batch_alter_table('mission_invitations', schema=None) as batch_op:
        batch_op.add_column(sa.Column('callsign', sa.String(length=255), nullable=True))
        batch_op.add_column(sa.Column('username', sa.String(length=255), nullable=True))
        batch_op.add_column(sa.Column('group_name', sa.String(length=255), nullable=True))
        batch_op.add_column(sa.Column('team_name', sa.String(length=255), nullable=True))
        batch_op.alter_column('client_uid',
               existing_type=sa.String(length=255),
               nullable=True)
        batch_op.create_foreign_key('mission_invitation_user', 'user', ['username'], ['username'])
        # batch_op.create_foreign_key('groups', 'groups', ['group_name'], ['group_name'])
        batch_op.create_foreign_key('mission_invitation_callsign', 'euds', ['callsign'], ['callsign'])
        batch_op.create_foreign_key('mission_invitation_team', 'teams', ['team_name'], ['name'])
        batch_op.create_foreign_key('mission_invitation_eud', 'euds', ['client_uid'], ['uid'])

    with op.batch_alter_table('missions', schema=None) as batch_op:
        batch_op.add_column(sa.Column('group', sa.String(length=255), nullable=True))
        # batch_op.create_foreign_key('groups', 'groups', ['group'], ['group_name'])
        batch_op.drop_column('group_name')

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('missions', schema=None) as batch_op:
        batch_op.add_column(sa.Column('group_name', sa.String(length=255), nullable=True))
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.drop_column('group')

    with op.batch_alter_table('mission_invitations', schema=None) as batch_op:
        batch_op.drop_constraint('user', type_='foreignkey')
        batch_op.drop_constraint('eud_callsign', type_='foreignkey')
        batch_op.drop_constraint('eud_uid', type_='foreignkey')
        batch_op.drop_constraint('teams', type_='foreignkey')
        batch_op.alter_column('client_uid',
               existing_type=sa.String(length=255),
               nullable=False)
        batch_op.drop_column('team_name')
        batch_op.drop_column('group_name')
        batch_op.drop_column('username')
        batch_op.drop_column('callsign')

    with op.batch_alter_table('mission_content', schema=None) as batch_op:
        batch_op.add_column(sa.Column('name', sa.String(length=255), nullable=True))
        batch_op.drop_column('filename')

    with op.batch_alter_table('mission_changes', schema=None) as batch_op:
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.drop_column('content_uid')

    # ### end Alembic commands ###



import os
import platform
import sys
from logging.handlers import TimedRotatingFileHandler

import flask_wtf
import sqlalchemy
import yaml
from flask_security import SQLAlchemyUserDatastore, Security
from flask_security.models import fsqla

from opentakserver.EmailValidator import EmailValidator
from opentakserver.PasswordValidator import PasswordValidator
# These unused imports are required by SQLAlchemy, don't remove them
from opentakserver.eud_handler.SocketServer import SocketServer
from opentakserver.models.EUD import EUD
from opentakserver.models.CoT import CoT
from opentakserver.models.Point import Point
from opentakserver.models.Alert import Alert
from opentakserver.models.DataPackage import DataPackage
from opentakserver.models.Certificate import Certificate
from opentakserver.models.Marker import Marker
from opentakserver.models.RBLine import RBLine
from opentakserver.models.Team import Team
from opentakserver.models.Group import Group
from opentakserver.models.EUDStats import EUDStats
from opentakserver.models.Mission import Mission
from opentakserver.models.MissionInvitation import MissionInvitation
from opentakserver.models.MissionContentMission import MissionContentMission
from opentakserver.models.MissionLogEntry import MissionLogEntry
from opentakserver.models.MissionChange import MissionChange
from opentakserver.models.MissionUID import MissionUID
from opentakserver.models.CasEvac import CasEvac
from opentakserver.models.ZMIST import ZMIST
from opentakserver.models.Chatrooms import Chatroom
from opentakserver.models.ChatroomsUids import ChatroomsUids
from opentakserver.models.VideoStream import VideoStream
from opentakserver.models.VideoRecording import VideoRecording
from opentakserver.models.WebAuthn import WebAuthn
from opentakserver.models.GroupMission import GroupMission
from opentakserver.extensions import db, logger, ldap_manager
from opentakserver.defaultconfig import DefaultConfig
import colorlog
from flask import Flask, jsonify
import logging
import argparse


def args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--ssl", help="Enable SSL", default=False, action=argparse.BooleanOptionalAction)
    return parser.parse_args()


def setup_logging(app):
    level = logging.INFO
    if app.config.get("DEBUG"):
        level = logging.DEBUG
    logger.setLevel(level)

    if sys.stdout.isatty():
        color_log_handler = colorlog.StreamHandler()
        color_log_formatter = colorlog.ColoredFormatter(
            '%(log_color)s[%(asctime)s] - eud_handler[%(process)d] - %(module)s - %(funcName)s - %(lineno)d - %(levelname)s - %(message)s',
            datefmt="%Y-%m-%d %H:%M:%S")
        color_log_handler.setFormatter(color_log_formatter)
        logger.addHandler(color_log_handler)

    os.makedirs(os.path.join(app.config.get("OTS_DATA_FOLDER"), "logs"), exist_ok=True)
    fh = TimedRotatingFileHandler(os.path.join(app.config.get("OTS_DATA_FOLDER"), 'logs', 'opentakserver.log'),
                                  when=app.config.get("OTS_LOG_ROTATE_WHEN"),
                                  interval=app.config.get("OTS_LOG_ROTATE_INTERVAL"),
                                  backupCount=app.config.get("OTS_BACKUP_COUNT"))
    fh.setFormatter(logging.Formatter(
        "[%(asctime)s] - eud_handler[%(process)d] - %(module)s - %(funcName)s - %(lineno)d - %(levelname)s - %(message)s"))
    logger.addHandler(fh)


def create_app():
    app = Flask(__name__)
    app.config.from_object(DefaultConfig)

    # Load config.yml if it exists
    if os.path.exists(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml")):
        app.config.from_file(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml"), load=yaml.safe_load)
    else:
        # First run, created config.yml based on default settings
        logger.info("Creating config.yml")
        with open(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml"), "w") as config:
            conf = {}
            for option in DefaultConfig.__dict__:
                # Fix the sqlite DB path on Windows
                if option == "SQLALCHEMY_DATABASE_URI" and platform.system() == "Windows" and DefaultConfig.__dict__[option].startswith("sqlite"):
                    conf[option] = DefaultConfig.__dict__[option].replace("////", "///").replace("\\", "/")
                elif option.isupper():
                    conf[option] = DefaultConfig.__dict__[option]
            config.write(yaml.safe_dump(conf))

    setup_logging(app)
    db.init_app(app)

    if app.config.get("OTS_ENABLE_LDAP"):
        logger.info("Enabling LDAP")
        ldap_manager.init_app(app)

    # The rest is required by flask, leave it in
    try:
        fsqla.FsModels.set_db_info(db)
    except sqlalchemy.exc.InvalidRequestError:
        pass

    from opentakserver.models.user import User
    from opentakserver.models.role import Role

    flask_wtf.CSRFProtect(app)
    user_datastore = SQLAlchemyUserDatastore(db, User, Role)
    app.security = Security(app, user_datastore, mail_util_cls=EmailValidator, password_util_cls=PasswordValidator)

    return app


app = create_app()

@app.route("/status")
def status():
    return jsonify({"status": "ok"})


def main():
    opts = args()
    if opts.ssl:
        socket_server = SocketServer(logger, app.app_context(), app.config.get("OTS_SSL_STREAMING_PORT"), True)
        logger.info(f"Started SSL server on port {app.config.get('OTS_SSL_STREAMING_PORT')}")
    else:
        socket_server = SocketServer(logger, app.app_context(), app.config.get("OTS_TCP_STREAMING_PORT"))
        logger.info(f"Started TCP server on port {app.config.get('OTS_TCP_STREAMING_PORT')}")
    socket_server.run()


if __name__ == "__main__":
    main()



import traceback

import bleach
import sqlalchemy
from flask import current_app as app, request, Blueprint, jsonify
from flask_babel import gettext
from flask_security import roles_accepted, hash_password, current_user, admin_change_password, auth_required

from opentakserver.blueprints.ots_api.api import search, paginate
from opentakserver.extensions import logger, db, ldap_manager
from opentakserver.models.EUD import EUD
from opentakserver.models.Group import Group
from opentakserver.models.GroupUser import GroupUser
from opentakserver.models.user import User
from opentakserver.UsernameValidator import UsernameValidator

user_api_blueprint = Blueprint('user_api_blueprint', __name__)


@user_api_blueprint.route("/api/user/add", methods=['POST'])
@roles_accepted("administrator")
def create_user():
    if app.config.get("OTS_ENABLE_LDAP"):
        return jsonify({'success': False, 'error': gettext(u'LDAP is enabled, please use your LDAP server to create users')}), 400

    username = bleach.clean(request.json.get('username'))
    password = bleach.clean(request.json.get('password'))
    confirm_password = bleach.clean(request.json.get('confirm_password'))

    validated_username = UsernameValidator(app).validate(username)
    if validated_username[0]:
        return jsonify({"success": False, "error": f"{validated_username[0]}"}), 400

    if password != confirm_password:
        return jsonify({'success': False, 'error': gettext(u'Passwords do not match')}), 400

    roles = request.json.get("roles")
    roles_cleaned = []

    for role in roles:
        role = bleach.clean(role)
        role_exists = app.security.datastore.find_role(role)

        if not role_exists:
            return jsonify({'success': False, 'error': gettext(u'Role %(role)s does not exist', role=role)}), 400

        elif role == 'administrator' and not current_user.has_role('administrator'):
            return jsonify({'success': False, 'error': gettext(u'Only administrators can add users to the administrators role')}), 403

        elif role not in roles_cleaned:
            roles_cleaned.append(role)

    if not app.security.datastore.find_user(username=username):
        logger.info("Creating user {}".format(username))
        app.security.datastore.create_user(username=username, password=hash_password(password), roles=roles_cleaned)
        db.session.commit()
        return jsonify({'success': True}), 200
    else:
        logger.error("User {} already exists".format(username))
        return jsonify({'success': False, 'error': gettext(u'User %(username)s already exists', username=username)}), 400


@user_api_blueprint.route("/api/user/delete", methods=['POST'])
@roles_accepted("administrator")
def delete_user():
    if app.config.get("OTS_ENABLE_LDAP"):
        return jsonify({'success': False, 'error': gettext(u'LDAP is enabled, please use your LDAP server to delete users')}), 400

    username = bleach.clean(request.json.get('username'))

    if username == current_user.username:
        return jsonify({'success': False, 'error': gettext(u"You can't delete your own account")}), 400

    logger.info("Deleting user {}".format(username))

    try:
        user = app.security.datastore.find_user(username=username)
        db.session.execute(sqlalchemy.delete(GroupUser).where(GroupUser.user_id == user.id))
        app.security.datastore.delete_user(user)
    except BaseException as e:
        logger.error(traceback.format_exc())
        return jsonify({'success': False, 'error': gettext(u'Failed to delete user: %(e)s', e=str(e))}), 400

    db.session.commit()
    return jsonify({'success': True}), 200


@user_api_blueprint.route("/api/user/password/reset", methods=['POST'])
@roles_accepted("administrator")
def admin_reset_password():
    if app.config.get("OTS_ENABLE_LDAP"):
        return jsonify({'success': False, 'error': gettext(u'LDAP is enabled, please use your LDAP server to reset passwords')}), 400

    username = bleach.clean(request.json.get("username"))
    new_password = bleach.clean(request.json.get("new_password"))

    if not username or not new_password:
        return jsonify({'success': False, 'error': gettext(u'Please specify a username and new password')}), 400

    if len(new_password) < app.config.get("SECURITY_PASSWORD_LENGTH_MIN"):
        return jsonify({'success': False, 'error': gettext(u'Your password must be at least %(characters)s characters long', characters=app.config.get("SECURITY_PASSWORD_LENGTH_MIN"))}), 400

    user = app.security.datastore.find_user(username=username)
    if user:
        admin_change_password(user, new_password, False)
        db.session.commit()
        return jsonify({'success': True}), 200
    else:
        return jsonify({'success': False, 'error': gettext(u'Could not find user %(username)s', username=username)}), 400


@user_api_blueprint.route('/api/user/deactivate', methods=['POST'])
@roles_accepted('administrator')
def deactivate_user():
    if app.config.get("OTS_ENABLE_LDAP"):
        return jsonify({'success': False, 'error': gettext(u'LDAP is enabled, please use your LDAP server to deactivate users')}), 400

    username = bleach.clean(request.json.get("username", ""))
    if not username:
        return jsonify({'success': False, 'error': gettext(u'Please specify the username to deactivate')}), 400

    user = app.security.datastore.find_user(username=username)
    if not user:
        return jsonify({'success': False, 'error': gettext(u'User %(username)s does not exist', username=username)}), 400

    deactivated = app.security.datastore.deactivate_user(user)
    if deactivated:
        db.session.commit()
        return jsonify({'success': True})
    else:
        return jsonify({'success': False, 'error': gettext('%(username)s is already deactivated', username=username)})


@user_api_blueprint.route('/api/user/activate', methods=['POST'])
@roles_accepted('administrator')
def activate_user():
    if app.config.get("OTS_ENABLE_LDAP"):
        return jsonify({'success': False, 'error': gettext(u'LDAP is enabled, please use your LDAP server to activate users')}), 400

    username = bleach.clean(request.json.get("username", ""))
    if not username:
        return jsonify({'success': False, 'error': gettext(u'Please specify the username to activate')}), 400

    user = app.security.datastore.find_user(username=username)
    if not user:
        return jsonify({'success': False, 'error': gettext(u'User %(username)s does not exist', username=username)}), 400

    activated = app.security.datastore.activate_user(user)
    if activated:
        db.session.commit()
        return jsonify({'success': True})
    else:
        return jsonify({'success': False, 'error': gettext(u'%(username)s is already activated', username=username)})


@user_api_blueprint.route("/api/user/role", methods=['POST'])
@roles_accepted("administrator")
def set_user_role():
    if app.config.get("OTS_ENABLE_LDAP"):
        return jsonify({'success': False, 'error': gettext(u'LDAP is enabled, please use your LDAP server to assign roles')}), 400

    username = bleach.clean(request.json.get("username", ""))
    roles = request.json.get("roles")
    roles_cleaned = []

    if not username or not roles:
        return jsonify({'success': False, 'error': gettext(u'Please specify a username and roles')}), 400

    for role in roles:
        role = bleach.clean(role)
        role_exists = app.security.datastore.find_role(role)

        if not role_exists:
            return jsonify({'success': False, 'error': gettext(u'Role %(role)s does not exist', role=role)}), 400

        elif role not in roles_cleaned:
            roles_cleaned.append(role)

    user = app.security.datastore.find_user(username=username)
    if not user:
        return jsonify({'success': False, 'error': gettext(u'User %(username)s does not exist', username=username)}), 400

    for role in user.roles:
        app.security.datastore.remove_role_from_user(user, role)

    for role in roles_cleaned:
        app.security.datastore.add_role_to_user(user, role)

    db.session.commit()
    return jsonify({'success': True})


@user_api_blueprint.route('/api/user/assign_eud', methods=['POST'])
@auth_required()
def assign_eud_to_user():
    username = bleach.clean(request.json.get('username')) if 'username' in request.json else None
    eud_uid = bleach.clean(request.json.get('uid')) if 'uid' in request.json else None
    user = None

    if not eud_uid:
        return {'success': False, 'error': 'Please specify an EUD'}, 400, {'Content-Type': 'application/json'}
    if not username or username == current_user.username:
        user = current_user
    elif username != current_user.username and current_user.has_role('administrator'):
        user = app.security.datastore.find_user(username=username)
        if not user:
            return jsonify({'success': False, 'error': gettext(u'User %(username)s does not exist', username=username)}), 404

    eud = db.session.query(EUD).filter_by(uid=eud_uid).first()

    if not eud:
        return jsonify({'success': False, 'error': gettext(u'EUD %(eud_uid)s not found', eud_uid=eud_uid)}), 404
    elif eud.user_id and not current_user.has_role('administrator') and current_user.id != eud.user_id:
        return jsonify({'success': False, 'error': gettext(u'%(uid)s is already assigned to another user', uid=eud.uid)}), 403
    else:
        eud.user_id = user.id
        db.session.add(eud)
        db.session.commit()

        return jsonify({'success': True})


@user_api_blueprint.route('/api/users')
@roles_accepted('administrator')
def get_users():
    if app.config.get("OTS_ENABLE_LDAP"):
        return jsonify({'success': False, 'error': gettext(u'LDAP is enabled, please use your LDAP server to manage users')}), 400

    query = db.session.query(User)
    query = search(query, User, 'username')

    return paginate(query)


@user_api_blueprint.route('/api/users/all')
@roles_accepted('administrator')
def get_all_users():
    if app.config.get("OTS_ENABLE_LDAP"):
        return jsonify({'success': False, 'error': gettext(u'LDAP is enabled, please use your LDAP server to manage users')}), 400

    users = db.session.execute(db.session.query(User)).all()
    return_value = []

    for user in users:
        user = user[0]
        return_value.append(user.serialize())

    return return_value


@user_api_blueprint.route('/api/users/groups')
@roles_accepted('administrator')
def get_user_groups():
    """ Gets a list of group memberships for a user
    :parameter: username

    :return: List of group memberships
    """
    if app.config.get("OTS_ENABLE_LDAP"):
        return jsonify({'success': False, 'error': gettext(u'LDAP is enabled, please use your LDAP server to manage groups')}), 400

    username = request.args.get("username")
    if not username:
        return jsonify({"success": False, "error": gettext(u"Please provide a username")}), 400

    username = bleach.clean(username)

    user = app.security.datastore.find_user(username=username)
    if not user:
        return jsonify({"success": False, "error": gettext(u"User %(username)s not found", username=username)}), 404

    group_memberships = db.session.execute(db.session.query(GroupUser).filter_by(user_id=user.id)).all()
    memberships = []
    for membership in group_memberships:
        membership: GroupUser = membership[0]
        memberships.append({"group_name": membership.group.name, "direction": membership.direction, "active": membership.enabled})

    return jsonify({"success": True, "results": memberships})


@user_api_blueprint.route('/api/users/groups', methods=["PUT"])
@roles_accepted("administrator")
def add_user_to_groups():
    """ Adds a user to one or more groups
    :parameter: groups - List of groups to add a user to
    :parameter: username
    :parameter: direction - Group direction, must be either IN or OUT

    :return: 400 if LDAP is enabled, no group or username is specified, or if the specified group or user doesn't exist or the user is already in the group. 200 on success.
    :rtype: Response
    """

    if app.config.get("OTS_ENABLE_LDAP"):
        return jsonify({'success': False, 'error': gettext(u'LDAP is enabled, please use your LDAP server to add users to groups')}), 400

    groups = request.json.get("groups")
    username = request.json.get("username")
    direction = request.json.get("direction")

    if not groups or not username or not direction:
        return jsonify({"success": False, "error": gettext(u"Please provide a list of groups, a username, and a direction")}), 400

    if direction != "IN" and direction != "OUT":
        return jsonify({"success": False, "error": gettext(u"Direction must be IN or OUT")}), 400

    user = app.security.datastore.find_user(username=username)
    if not user:
        return jsonify({"success": False, "error": gettext(u"User %(username)s doesn't exist", username=username)}), 404

    for group_name in groups:
        group_name = bleach.clean(group_name)
        group = db.session.execute(db.session.query(Group).filter_by(name=group_name)).first()
        if not group:
            return jsonify({"success": False, "error": gettext(u"Group %(group_name)s doesn't exist", group_name=group_name)}), 404

        group = group[0]

        membership = GroupUser()
        membership.user_id = user.id
        membership.group_id = group.id
        membership.direction = direction

        try:
            db.session.add(membership)
            db.session.commit()
        except sqlalchemy.exc.IntegrityError:
            db.session.rollback()

    return jsonify({"success": True})



"""Set install on enrollment and connection to nullable

Revision ID: 6af2256c568d
Revises: 34d5df698b81
Create Date: 2024-09-20 20:17:28.926142

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '6af2256c568d'
down_revision = '34d5df698b81'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('data_packages', schema=None) as batch_op:
        batch_op.alter_column('install_on_enrollment',
               existing_type=sa.BOOLEAN(),
               nullable=True)
        batch_op.alter_column('install_on_connection',
               existing_type=sa.BOOLEAN(),
               nullable=True)

    with op.batch_alter_table('device_profiles', schema=None) as batch_op:
        batch_op.alter_column('enrollment',
               existing_type=sa.BOOLEAN(),
               nullable=True)
        batch_op.alter_column('connection',
               existing_type=sa.BOOLEAN(),
               nullable=True)
        try:
            batch_op.drop_column('id')
        except BaseException as e:
            print(f"Failed to drop euds.id column: {e}")
            pass

    with op.batch_alter_table('packages', schema=None) as batch_op:
        batch_op.alter_column('install_on_enrollment',
               existing_type=sa.BOOLEAN(),
               nullable=True)
        batch_op.alter_column('install_on_connection',
               existing_type=sa.BOOLEAN(),
               nullable=True)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('packages', schema=None) as batch_op:
        batch_op.alter_column('install_on_connection',
               existing_type=sa.BOOLEAN(),
               nullable=False)
        batch_op.alter_column('install_on_enrollment',
               existing_type=sa.BOOLEAN(),
               nullable=False)

    with op.batch_alter_table('device_profiles', schema=None) as batch_op:
        batch_op.add_column(sa.Column('id', sa.INTEGER(), nullable=False))
        batch_op.alter_column('connection',
               existing_type=sa.BOOLEAN(),
               nullable=False)
        batch_op.alter_column('enrollment',
               existing_type=sa.BOOLEAN(),
               nullable=False)

    with op.batch_alter_table('data_packages', schema=None) as batch_op:
        batch_op.alter_column('install_on_connection',
               existing_type=sa.BOOLEAN(),
               nullable=False)
        batch_op.alter_column('install_on_enrollment',
               existing_type=sa.BOOLEAN(),
               nullable=False)

    # ### end Alembic commands ###



import datetime
import uuid
from dataclasses import dataclass

from opentakserver.functions import iso8601_string_from_datetime
from opentakserver.extensions import db
from sqlalchemy import Integer, String, DateTime, JSON
from sqlalchemy.orm import Mapped, mapped_column, relationship


@dataclass
class MissionContent(db.Model):
    __tablename__ = "mission_content"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    keywords: Mapped[JSON] = mapped_column(JSON, nullable=True)
    mime_type: Mapped[str] = mapped_column(String(255), nullable=True)
    filename: Mapped[str] = mapped_column(String(255), nullable=True)
    submission_time: Mapped[datetime] = mapped_column(DateTime, nullable=True)
    submitter: Mapped[str] = mapped_column(String(255), nullable=True)
    uid: Mapped[str] = mapped_column(String(255), unique=True, default=str(uuid.uuid4()))
    creator_uid: Mapped[str] = mapped_column(String(255), nullable=True)
    hash: Mapped[str] = mapped_column(String(255), nullable=False, unique=True)
    size: Mapped[int] = mapped_column(Integer, nullable=True)
    expiration: Mapped[int] = mapped_column(Integer, nullable=True)
    mission_changes = relationship("MissionChange", back_populates="content_resource")
    mission = relationship("Mission", secondary="mission_content_mission", back_populates="contents", cascade="all, delete")

    def serialize(self):
        return {
            "keywords": self.keywords,
            "mime_type": self.mime_type,
            "filename": self.filename,
            "submission_time": self.submission_time,
            "submitter": self.submitter,
            "uid": self.uid,
            "creator_uid": self.creator_uid,
            "hash": self.hash,
            "size": self.size,
            "expiration": self.expiration
        }

    def to_json(self):
        return {
            "data": {
                "keywords": self.keywords if self.keywords else [],
                "mimeType": self.mime_type,
                "name": self.filename,
                "submissionTime": iso8601_string_from_datetime(self.submission_time),
                "submitter": self.submitter,
                "uid": self.uid,
                "creator_uid": self.creator_uid,
                "hash": self.hash,
                "size": self.size,
                "expiration": self.expiration
            },
            "timestamp": iso8601_string_from_datetime(self.submission_time),
            "creatorUid": self.creator_uid
        }



import hashlib
import os
import traceback

from flask import Blueprint, jsonify, current_app as app, request
from flask_babel import gettext
from flask_security import roles_required

from opentakserver.extensions import logger
from werkzeug.utils import secure_filename

plugin_blueprint = Blueprint("plugin_api_blueprint", __name__)


@plugin_blueprint.route("/api/plugins")
@roles_required("administrator")
def get_plugins():
    if hasattr(app, 'plugin_manager'):
        plugins = []

        for plugin in app.plugin_manager.plugins.values():
            plugins.append(plugin.get_info())

        return jsonify({'success': True, 'plugins': plugins})
    else:
        return jsonify({'success': True, 'plugins': []}), 200


@plugin_blueprint.route("/api/plugins", methods=["POST"])
@roles_required("administrator")
def upload_plugin():
    if hasattr(app, 'plugin_manager'):
        if not len(request.files):
            return jsonify({'error': 'You must upload a plugin file'}), 400
        for file in request.files:
            try:
                file = request.files[file]

                sanitized_filename = secure_filename(file.filename)
                name, extension = os.path.splitext(sanitized_filename)
                extension = extension.replace(".", "")

                if extension not in ["zip", "whl", "gz"]:
                    return jsonify({"success": False, "error": gettext(u"Server plugins must be zip, whl, or tar.gz files")}), 400

                file.save(os.path.join(app.config.get("UPLOAD_FOLDER"), sanitized_filename))

                return jsonify({"success": True})
            except BaseException as e:
                logger.error(f"Failed to save plugin file: {e}")
                logger.debug(traceback.format_exc())
                return jsonify({"success": False, "error": e}), 500
    else:
        return jsonify({'success': False, 'error': gettext(u"Plugins are disabled")}), 400


@plugin_blueprint.route("/api/plugins/repo")
@roles_required("administrator")
def get_plugin_repo():
    return jsonify({"success": True, "repo_url": app.config.get("OTS_PLUGIN_REPO")})


@plugin_blueprint.route("/api/plugins/<plugin_name>", strict_slashes=False)
@roles_required("administrator")
def get_plugin(plugin_name: str):
    if hasattr(app, 'plugin_manager'):
        plugin = app.plugin_manager.plugins.get(plugin_name)
        if plugin:
            plugin_metadata = plugin.load_metadata()
            plugin_metadata['enabled'] = app.plugin_manager.check_if_plugin_enabled(plugin.name.lower())
            return jsonify(plugin_metadata)
        else:
            return jsonify({'success': False, 'error': gettext(u'Plugin %(plugin_name)s not found', plugin_name=plugin_name)}), 404
    else:
        return jsonify({'success': False, 'error': gettext(u'Plugins are disabled')}), 400


@plugin_blueprint.route("/api/plugins/<plugin_name>/disable", methods=["POST"])
@roles_required("administrator")
def disable_plugin(plugin_name):
    if hasattr(app, 'plugin_manager'):
        try:
            logger.info(f"Disabling plugin {plugin_name}")
            app.plugin_manager.disable_plugin(plugin_name)
            return jsonify({'success': True})
        except BaseException as e:
            logger.error(f"Failed to disable {plugin_name}: {e}")
            logger.error(traceback.format_exc())
            return jsonify({'success': False, 'error': str(e)})
    else:
        return jsonify({'success': False, 'error': gettext(u'Plugins are disabled')}), 400


@plugin_blueprint.route("/api/plugins/<plugin_name>/enable", methods=["POST"])
@roles_required("administrator")
def enable_plugin(plugin_name):
    if hasattr(app, 'plugin_manager'):
        try:
            app.plugin_manager.enable_plugin(plugin_name)
            return jsonify({'success': True})
        except BaseException as e:
            logger.error(f"Failed to enable {plugin_name}: {e}")
            logger.error(traceback.format_exc())
            return jsonify({'success': False, 'error': str(e)})
    else:
        return jsonify({'success': False, 'error': gettext(u'Plugins are disabled')}), 400



import datetime
import json
import os
import pathlib
import traceback
import uuid
from urllib.parse import urlparse

from ffmpeg import FFmpeg
from flask_babel import gettext
from flask_ldap3_login import AuthenticationResponseStatus
from sqlalchemy import update
from werkzeug.datastructures import ImmutableMultiDict

import bleach
import requests
import sqlalchemy.exc
from flask import current_app as app, request, Blueprint, jsonify
from flask_security import auth_required, current_user, verify_password
from flask_security.utils import parse_auth_token

from opentakserver.extensions import logger, db, ldap_manager
from opentakserver.models.VideoStream import VideoStream
from opentakserver.forms.MediaMTXPathConfig import MediaMTXPathConfig
from opentakserver.models.VideoRecording import VideoRecording

mediamtx_api_blueprint = Blueprint('mediamtx_api_blueprint', __name__)


def get_stream_protocol(source_type):
    protocol = "rtsp"
    if source_type.startswith('rtsps'):
        protocol = 'rtsps'
    if source_type.startswith('rtsp'):
        protocol = "rtsp"
    elif source_type == 'hlsSource':
        protocol = "hls"
    elif source_type == 'rpiCameraSource':
        protocol = "rpi_camera"
    elif source_type.startswith('rtmp'):
        protocol = 'rtmp'
    elif source_type.startswith('srt'):
        protocol = 'srt'
    elif source_type.startswith('udp'):
        protocol = 'udp'
    elif source_type.startswith('webRTC'):
        protocol = 'webrtc'

    return protocol


@mediamtx_api_blueprint.route('/api/mediamtx/webhook')
def mediamtx_webhook():
    token = request.args.get('token')
    if not token or bleach.clean(token) != app.config.get("OTS_MEDIAMTX_TOKEN"):
        logger.error('Invalid token')
        return jsonify({'success': False, 'error': gettext(u'Invalid token')}), 401

    event = bleach.clean(request.args.get('event'))
    if event == 'init':
        rtsp_port = bleach.clean(request.args.get("rtsp_port"))
        path = bleach.clean(request.args.get("path"))

        if path == 'startup':
            paths = VideoStream.query.all()
            for path in paths:
                r = requests.post(
                    "{}/v3/config/paths/add/{}".format(app.config.get("OTS_MEDIAMTX_API_ADDRESS"), path.path),
                    json=json.loads(path.mediamtx_settings))
                logger.debug("Init added {} {}".format(path, r.status_code))

            # Get all paths from MediaMTX and make sure they're in OTS's database
            r = requests.get("{}/v3/paths/list".format(app.config.get("OTS_MEDIAMTX_API_ADDRESS")))
            paths = r.json()
            for path in paths['items']:
                video_stream = db.session.query(VideoStream).where(VideoStream.path == path['name']).first()
                if not video_stream:
                    if not path['source']:
                        continue
                    video_stream = VideoStream()
                    video_stream.protocol = get_stream_protocol(path['source']['type'])

                    r = requests.get("{}/v3/config/global/get".format(app.config.get("OTS_MEDIAMTX_API_ADDRESS")))
                    video_stream.port = r.json()['rtspAddress'].replace(":", "")

                    r = requests.get(
                        "{}/v3/config/paths/get/{}".format(app.config.get("OTS_MEDIAMTX_API_ADDRESS"), path['name']))
                    video_stream.mediamtx_settings = json.dumps(r.json())

                    video_stream.path = path['name']
                    video_stream.alias = path['name']
                    video_stream.rtsp_reliable = 1
                    video_stream.ready = event == 'ready'
                    video_stream.rover_port = -1
                    video_stream.ignore_embedded_klv = False
                    video_stream.buffer_time = None
                    video_stream.network_timeout = 10000
                    video_stream.uid = str(uuid.uuid4())
                    video_stream.generate_xml(urlparse(request.url_root).hostname)

                    db.session.add(video_stream)
                    db.session.commit()

    elif event == 'connect':
        connection_type = bleach.clean(request.args.get("connection_type"))
        connection_id = bleach.clean(request.args.get("connection_id"))
        rtsp_port = bleach.clean(request.args.get("rtsp_port"))
    elif event == 'ready' or event == 'notready':
        rtsp_port = bleach.clean(request.args.get("rtsp_port"))
        path = bleach.clean(request.args.get("path"))
        query = bleach.clean(request.args.get("query"))
        source_type = bleach.clean(request.args.get("source_type"))
        source_id = bleach.clean(request.args.get("source_id"))

        video_stream = db.session.query(VideoStream).where(VideoStream.path == path).first()
        if video_stream:
            video_stream.ready = event == 'ready'
            db.session.add(video_stream)
            db.session.commit()
            r = requests.patch("{}/v3/config/paths/patch/{}".format(app.config.get("OTS_MEDIAMTX_API_ADDRESS"), path),
                               json=json.loads(video_stream.mediamtx_settings))
            logger.debug("Ready Patched path {}: {} - {}".format(path, r.status_code, r.text))
        else:
            video_stream = VideoStream()
            if source_type.startswith('rtsps'):
                video_stream.protocol = 'rtsps'
            if source_type.startswith('rtsp'):
                video_stream.protocol = "rtsp"
            elif source_type == 'hlsSource':
                video_stream.protocol = "hls"
            elif source_type == 'rpiCameraSource':
                video_stream.protocol = "rpi_camera"
            elif source_type.startswith('rtmp'):
                video_stream.protocol = 'rtmp'
            elif source_type.startswith('srt'):
                video_stream.protocol = 'srt'
            elif source_type.startswith('udp'):
                video_stream.protocol = 'udp'
            elif source_type.startswith('webRTC'):
                video_stream.protocol = 'webrtc'

            # video_stream.query = query
            video_stream.port = rtsp_port
            video_stream.path = path
            video_stream.alias = path
            video_stream.rtsp_reliable = 1
            video_stream.ready = event == 'ready'
            video_stream.rover_port = -1
            video_stream.ignore_embedded_klv = False
            video_stream.buffer_time = None
            video_stream.network_timeout = 10000
            video_stream.uid = str(uuid.uuid4())
            video_stream.generate_xml(urlparse(request.url_root).hostname)
            mediamtx_settings = MediaMTXPathConfig(None)
            mediamtx_settings.sourceOnDemand.data = source_id is not None
            mediamtx_settings.record.data = False
            video_stream.mediamtx_settings = json.dumps(mediamtx_settings.serialize())

            db.session.add(video_stream)
            db.session.commit()

        if event == 'ready':
            os.makedirs(os.path.join(app.config.get('OTS_DATA_FOLDER'), "mediamtx", "recordings", video_stream.path),
                        exist_ok=True)

            try:
                (FFmpeg().input(
                    video_stream.to_json()['rtsp_link'] + "?token={}".format(token))
                 .option("y")
                 .output(os.path.join(app.config.get('OTS_DATA_FOLDER'), "mediamtx", "recordings", video_stream.path,
                                      "thumbnail.png"), {"frames:v": 1}).execute())
            except BaseException as e:
                logger.error(f"Failed to create thumbnail: {e}")
                logger.debug(traceback.format_exc())

    elif event == 'read':
        rtsp_port = bleach.clean(request.args.get("rtsp_port"))
        path = bleach.clean(request.args.get("path"))
        query = bleach.clean(request.args.get("query"))
        reader_type = bleach.clean(request.args.get("reader_type"))
        reader_id = bleach.clean(request.args.get("reader_id"))
    elif event == 'disconnect':
        connection_type = bleach.clean(request.args.get("connection_type"))
        connection_id = bleach.clean(request.args.get("connection_id"))
        rtsp_port = bleach.clean(request.args.get("rtsp_port"))
    elif event == 'segment_record':
        recording = VideoRecording()
        recording.segment_path = bleach.clean(request.args.get('segment_path'))
        recording.path = bleach.clean(request.args.get('path'))
        recording.in_progress = True
        recording.start_time = datetime.datetime.now(datetime.timezone.utc)

        with (app.app_context()):
            try:
                db.session.add(recording)
                db.session.commit()
            except sqlalchemy.exc.IntegrityError:
                db.session.rollback()
                db.session.execute(update(VideoRecording).filter(VideoRecording.segment_path == recording.segment_path)
                                   .values(**recording.serialize()))
                db.session.commit()
    elif event == 'segment_record_complete':
        segment_path = bleach.clean(request.args.get("segment_path"))
        with app.app_context():
            recording = db.session.execute(db.session.query(VideoRecording)
                                           .filter(VideoRecording.segment_path == segment_path)).first()
            if recording and recording.count:
                recording = recording[0]
                recording.in_progress = False
                recording.stop_time = datetime.datetime.now(datetime.timezone.utc)
                recording.duration = (recording.stop_time - recording.start_time.replace(tzinfo=datetime.timezone.utc)).seconds
            else:
                recording = VideoRecording()
                recording.segment_path = bleach.clean(request.args.get('segment_path'))
                recording.path = bleach.clean(request.args.get('path'))
                recording.in_progress = False
                recording.stop_time = datetime.datetime.now(datetime.timezone.utc)

            try:
                probe = json.loads(FFmpeg(executable="ffprobe").input(recording.segment_path, print_format="json", show_streams=None, show_format=None).execute())
                for stream in probe['streams']:
                    if stream['codec_type'].lower() == 'video':
                        recording.width = stream['width']
                        recording.height = stream['height']
                        recording.video_bitrate = stream['bit_rate']
                        recording.video_codec = stream['codec_name']
                        FFmpeg().input(recording.segment_path, ss="00:00:01").option("y").output(
                            recording.segment_path + ".png", {"frames:v": 1}).execute()
                    elif stream['codec_type'].lower() == 'audio':
                        recording.audio_codec = stream['codec_name']
                        recording.audio_samplerate = stream['sample_rate']
                        recording.audio_channels = stream['channels']
                        recording.audio_bitrate = stream['bit_rate']
                if 'format' in probe and 'size' in probe['format']:
                    recording.file_size = probe['format']['size']
            except BaseException as e:
                logger.error(f"Failed to run ffprobe: {e}")
                logger.debug(traceback.format_exc())

            db.session.add(recording)
            db.session.commit()
        pass

    return '', 200


@mediamtx_api_blueprint.route('/api/mediamtx/stream/add', methods=['POST'])
@mediamtx_api_blueprint.route('/api/mediamtx/stream/update', methods=['PATCH'])
@auth_required()
def add_update_stream():
    try:
        form = MediaMTXPathConfig(formdata=ImmutableMultiDict(request.json))
        if not form.validate():
            return jsonify({'success': False, 'errors': form.errors}), 400

        path = bleach.clean(request.json.get("path", ""))

        if not path:
            return jsonify({'success': False, 'error': gettext(u'Please specify a path name')}), 400

        if path.startswith("/"):
            return jsonify({'success': False, 'error': gettext(u'Path cannot begin with a slash')}), 400

        video = db.session.query(VideoStream).where(VideoStream.path == path).first()
        if not video and request.path.endswith('add'):
            video = VideoStream()
            video.path = path
            video.username = current_user.username
            video.mediamtx_settings = json.dumps(form.serialize())
            video.rover_port = -1
            video.ignore_embedded_klv = False
            video.buffer_time = None
            video.rtsp_reliable = 1
            video.network_timeout = 10000
            video.generate_xml(urlparse(request.url_root).hostname)
        elif not video and request.path.endswith('update'):
            return jsonify({'success': False, 'error': gettext(u'Path %(path)s not found', path=path)}), 400

        settings = json.loads(video.mediamtx_settings)

        for setting in request.json:
            if setting == 'csrf_token' or setting == 'sourceOnDemand' or setting == 'path':
                continue

            # When the source URL is blank set sourceOnDemand to false in order to avoid a MediaMTX error
            if setting == 'source':
                source = request.json.get(setting)
                if not source:
                    settings["sourceOnDemand"] = False
                # Run ffmpeg and yt-dlp for live YouTube videos
                elif "youtube.com" in source or "youtu.be" in source:
                    settings["source"] = None
                    settings["sourceOnDemand"] = False
                    yt_dlp_path = os.path.join(pathlib.Path.home(), ".opentakserver_venv", "bin", "yt-dlp")
                    settings["runOnDemand"] = f"sh -c 'ffmpeg -re -i \"$({yt_dlp_path} -g {source})\" -c:v copy -f rtsp rtsp://127.0.0.1:8554/$RTSP_PATH'"
                continue

            key = bleach.clean(setting)
            value = request.json.get(setting)
            if isinstance(value, str):
                value = bleach.clean(value)
            if value is not None:
                settings[key] = value
                logger.debug("set {} to {}".format(key, value))

        if request.path.endswith('update'):
            r = requests.patch("{}/v3/config/paths/patch/{}".format(app.config.get("OTS_MEDIAMTX_API_ADDRESS"), path),
                               json=settings)
        else:
            r = requests.post("{}/v3/config/paths/add/{}".format(app.config.get("OTS_MEDIAMTX_API_ADDRESS"), path),
                              json=settings)

        if r.status_code == 200:
            logger.debug("Patched path {}: {}".format(path, r.status_code))
            video.mediamtx_settings = json.dumps(settings)
            db.session.add(video)
            db.session.commit()
            return jsonify({'success': True})

        else:
            action = 'add' if request.path.endswith('add') else 'update'
            logger.error("Failed to {} mediamtx path: {} - {}".format(action, r.status_code, r.json()['error']))
            return jsonify({'success': False, 'error': r.json()['error']}), 400
    except BaseException as e:
        logger.error(traceback.format_exc())
        return jsonify({'success': False, 'error': str(e)}), 500


@mediamtx_api_blueprint.route('/api/mediamtx/stream/delete', methods=['DELETE'])
@auth_required()
def delete_stream():
    try:
        path = bleach.clean(request.args.get("path", ""))

        if not path:
            return jsonify({'success': False, 'error': gettext(u'Please specify a path name')}), 400

        r = requests.delete('{}/v3/config/paths/delete/{}'.format(app.config.get("OTS_MEDIAMTX_API_ADDRESS"), path))
        logger.debug("Delete status code: {}".format(r.status_code))
        video = db.session.query(VideoStream).filter(VideoStream.path == path)
        if not video:
            return jsonify({'success': False, 'error': gettext(u'Path %(path)s not found', path=path)}), 400

        video.delete()
        db.session.commit()
    except requests.exceptions.ConnectionError as e:
        logger.error(traceback.format_exc())
        return jsonify({'success': False, 'error': gettext(u'MediaMTX is not running')}), 500

    if r.status_code != 404:
        return r.text, r.status_code
    else:
        return "", 200


# This is mainly for mediamtx authentication
@mediamtx_api_blueprint.route('/api/external_auth', methods=['POST'])
def external_auth():
    username = bleach.clean(request.json.get('user'))
    password = bleach.clean(request.json.get('password'))
    action = bleach.clean(request.json.get('action'))
    query = bleach.clean(request.json.get('query'))
    ip = bleach.clean(request.json.get('ip'))

    # Whitelist 127.0.0.1 to make things like YouTube video re-streaming work
    if ip and ip in app.config.get("OTS_IP_WHITELIST"):
        return '', 200

    # Token auth to prevent high CPU usage when reading HLS streams
    if 'jwt' in query or 'token' in query:
        query = query.split("&")
        for q in query:
            if "=" not in q:
                continue
            key, value = q.split("=")
            if key == 'jwt':
                try:
                    parse_auth_token(value)
                    return '', 200
                except BaseException as e:
                    logger.error(f"Invalid token: {e}")
                    return '', 401
            elif key == 'token':
                if value == app.config.get("OTS_MEDIAMTX_TOKEN"):
                    return '', 200
                else:
                    return '', 401

    auth_success = False

    # LDAP Auth
    if app.config.get("OTS_ENABLE_LDAP"):
        result = ldap_manager.authenticate(username, password)
        if result.status == AuthenticationResponseStatus.success:
            # Keep this import here to avoid a circular import when OTS is started
            from opentakserver.blueprints.ots_api.ldap_api import save_user

            save_user(result.user_dn, result.user_id, result.user_info, result.user_groups)
            auth_success = True
        else:
            return '', 401
    # Flask-Security auth
    else:
        user = app.security.datastore.find_user(username=username)
        if not user:
            return '', 401
        if not verify_password(password, user.password):
            return '', 401
        auth_success = True

    if auth_success:
        if action == 'publish':
            logger.debug("Publish {}".format(request.json.get('path')))
            v = VideoStream()
            v.uid = bleach.clean(request.json.get('id')) if request.json.get('id') else None
            v.rover_port = -1
            v.ignore_embedded_klv = False
            v.buffer_time = None
            v.network_timeout = 10000
            v.protocol = bleach.clean(request.json.get('protocol'))
            v.path = bleach.clean(request.json.get('path'))
            v.alias = v.path.split("/")[-1]
            v.username = bleach.clean(request.json.get('user'))
            path_config = MediaMTXPathConfig(None).serialize()
            path_config['sourceOnDemand'] = False
            v.mediamtx_settings = json.dumps(path_config)

            if v.protocol == 'rtsp':
                v.port = 8554
                v.rtsp_reliable = 1
            elif v.protocol == 'rtmp':
                v.port = 1935
                v.rtsp_reliable = 1
            else:
                v.rtsp_reliable = 0

            v.generate_xml(request.json.get("ip"))

            with app.app_context():
                try:

                    db.session.add(v)
                    db.session.commit()
                    r = requests.post(
                        "{}/v3/config/paths/add/{}".format(app.config.get("OTS_MEDIAMTX_API_ADDRESS"), v.path),
                        json=path_config)
                    if r.status_code == 200:
                        logger.debug("Added path {} to mediamtx".format(v.path))
                    else:
                        logger.error(
                            "Failed to add path {} to mediamtx. Status code {} {}".format(v.path, r.status_code,
                                                                                          r.text))
                    logger.debug("Inserted video stream {}".format(v.uid))
                except sqlalchemy.exc.IntegrityError as e:
                    try:
                        db.session.rollback()
                        video = db.session.query(VideoStream).filter(VideoStream.path == v.path).first()
                        r = requests.post(
                            "{}/v3/config/paths/add/{}".format(app.config.get("OTS_MEDIAMTX_API_ADDRESS"), v.path),
                            json=json.loads(video.mediamtx_settings))
                        if r.status_code == 200:
                            logger.debug("Added path {} to mediamtx".format(v.path))
                        else:
                            logger.error(
                                "Failed to add path {} to mediamtx. Status code {} {}".format(v.path, r.status_code,
                                                                                              r.text))
                    except:
                        logger.error(traceback.format_exc())

        logger.debug("external_auth returning 200")
        return '', 200
    elif query:
        for arg in query.split("&"):
            key, value = arg.split("=")
            if key == 'token' and value == app.config.get("OTS_MEDIAMTX_TOKEN"):
                return '', 200
    else:
        logger.debug("external_auth returning 401")
        return '', 401



from flask import current_app as app, request, Blueprint
from flask_ldap3_login.forms import LDAPLoginForm
from flask_security import login_user, current_user
from flask_security.utils import base_render_json
from werkzeug.datastructures import ImmutableMultiDict

from opentakserver.extensions import ldap_manager
from opentakserver.models.user import User
from opentakserver.extensions import logger

ldap_blueprint = Blueprint('ldap_blueprint', __name__)


@ldap_manager.save_user
def save_user(dn: str, username: str, data, groups):
    """
    This callback will check Flask-Security's database for a user. If the user exists, all roles are removed. If
    the user doesn't exist, it is created. Finally, the LDAP groups the user belongs  are then added as Flask-Security roles.
    This allows admins to change user groups and have them work immediately on the next login.

    :param dn: User's LDAP Distinguished Name
    :param username: The Username
    :param data: All LDAP attributes for the user
    :param groups: A list of groups the user belongs to
    :return: User object
    """

    user: User = app.security.datastore.find_user(username=username)
    if user:
        roles = user.roles
        for role in roles:
            app.security.datastore.remove_role_from_user(user, role)
    else:
        user = app.security.datastore.create_user(username=username, password=None)

    is_admin = False
    for group in groups:
        if group['cn'] == app.config.get("OTS_LDAP_ADMIN_GROUP"):
            is_admin = True
        elif group['cn'].startswith(app.config.get("OTS_LDAP_PREFERENCE_ATTRIBUTE_PREFIX")) and not (group['cn'].lower().endswith("_read") or group['cn'].lower().endswith("_write")):
            logger.debug(f"Adding {group['cn']} role to {user.username}")
            app.security.datastore.add_role_to_user(user, app.security.datastore.find_or_create_role(group['cn']))

    if not is_admin:
        app.security.datastore.add_role_to_user(user, 'user')
    else:
        app.security.datastore.add_role_to_user(user, 'administrator')

    app.security.datastore.commit()
    return user


@ldap_blueprint.route("/api/ldap_login", methods=["POST"])
def ldap_login():
    form = LDAPLoginForm(formdata=ImmutableMultiDict(request.json))

    # LDAPLoginForm.validate() will call save_user()
    if form.validate():
        login_user(form.user, app.config.get("SECURITY_DEFAULT_REMEMBER_ME"), authn_via=["ldap"])
        payload = {"identity_attributes": {"ldap": {}}}
        return base_render_json(form, include_auth_token=True, additional=payload)

    else:
        return base_render_json(form)



import base64
import os
import datetime
import traceback
from urllib.parse import urlparse

from xml.etree.ElementTree import Element, tostring, SubElement

import bleach
import sqlalchemy
from OpenSSL import crypto
from flask import current_app as app, request, Blueprint, jsonify
from flask_ldap3_login import AuthenticationResponseStatus
from flask_security import verify_password

from opentakserver.extensions import logger, db, ldap_manager

from opentakserver.models.EUD import EUD
from opentakserver.models.Token import Token

from opentakserver.certificate_authority import CertificateAuthority

from opentakserver.models.Certificate import Certificate

certificate_authority_api_blueprint = Blueprint('certificate_authority_api_blueprint', __name__)


# flask-security's http_auth_required() decorator will deny access because ATAK doesn't do CSRF,
# so we handle basic auth ourselves
def basic_auth(credentials):
    try:
        username, password = base64.b64decode(credentials.split(" ", 1)[-1].encode('utf-8')).decode('utf-8').split(":", 1)
        username = bleach.clean(username)
        password = bleach.clean(password)

        if app.config.get("OTS_ENABLE_LDAP"):
            result = ldap_manager.authenticate(username, password)

            if result.status == AuthenticationResponseStatus.success:
                # Keep this import here to avoid a circular import when OTS is started
                from opentakserver.blueprints.ots_api.ldap_api import save_user

                save_user(result.user_dn, result.user_id, result.user_info, result.user_groups)
                return True
            return False

        user = app.security.datastore.find_user(username=username)
        if not user:
            logger.error(f"User {username} doesn't exist")
            return False
        elif verify_password(password, user.password):
            return True
        else:
            return Token.verify_token(password)

    except BaseException as e:
        logger.error("Failed to verify credentials: {}".format(e))
        return False


# require basic auth
@certificate_authority_api_blueprint.route('/Marti/api/tls/config')
def tls_config():
    root_element = Element('ns2:certificateConfig')
    root_element.set('xmlns', "http://bbn.com/marti/xml/config")
    root_element.set('xmlns:ns2', "com.bbn.marti.config")

    name_entries = SubElement(root_element, "nameEntries")
    first_name_entry = SubElement(name_entries, "nameEntry")
    first_name_entry.set('name', 'O')
    first_name_entry.set('value', app.config.get('OTS_CA_ORGANIZATION'))

    second_name_entry = SubElement(name_entries, "nameEntry")
    second_name_entry.set('name', 'OU')
    second_name_entry.set('value', app.config.get('OTS_CA_ORGANIZATIONAL_UNIT'))

    return tostring(root_element), 200, {'Content-Type': 'application/xml'}


@certificate_authority_api_blueprint.route('/Marti/api/tls/signClient/', methods=['POST'])
def sign_csr():
    if not basic_auth(request.headers.get('Authorization')):
        return '', 401
    return '', 200


@certificate_authority_api_blueprint.route('/Marti/api/tls/signClient/v2', methods=['POST'])
def sign_csr_v2():
    if not basic_auth(request.headers.get('Authorization')):
        return '', 401

    try:
        if 'clientUID' in request.args.keys():
            uid = request.args.get('clientUID')
        else:
            uid = request.args.get("clientUid")

        csr = request.data.decode('utf-8')
        if "BEGIN CERTIFICATE REQUEST" not in csr:
            csr = '-----BEGIN CERTIFICATE REQUEST-----\n' + csr

        if not csr.endswith('\n'):
            csr = csr + '\n'
            
        csr = csr + '-----END CERTIFICATE REQUEST-----'

        x509 = crypto.load_certificate_request(crypto.FILETYPE_PEM, csr.encode())

        common_name = x509.get_subject().CN
        logger.debug("Attempting to sign CSR for {}".format(common_name))

        cert_authority = CertificateAuthority(logger, app)

        signed_csr = cert_authority.sign_csr(csr.encode(), common_name, False).decode("utf-8")
        signed_csr = signed_csr.replace("-----BEGIN CERTIFICATE-----\n", "")
        signed_csr = signed_csr.replace("\n-----END CERTIFICATE-----\n", "")

        f = open(os.path.join(app.config.get("OTS_CA_FOLDER"), "ca.pem"), 'r')
        cert = f.read()
        f.close()

        cert = cert.replace("-----BEGIN CERTIFICATE-----\n", "")
        cert = cert.replace("\n-----END CERTIFICATE-----\n", "")

        # iTAK expects a JSON response but with the Content-Type header set to text/plain for some reason
        if request.headers.get('Accept') == 'text/plain' or request.headers.get('Accept') == 'application/json' or request.headers.get('Accept') == "*/*":
            response = {'signedCert': signed_csr, 'ca0': cert, 'ca1': cert}
        else:
            enrollment = Element('enrollment')
            signed_cert = SubElement(enrollment, 'signedCert')
            signed_cert.text = signed_csr
            ca = SubElement(enrollment, 'ca')
            ca.text = cert

            response = tostring(enrollment).decode('utf-8')
            response = '<?xml version="1.0" encoding="UTF-8"?>\n' + response

        username, password = base64.b64decode(
            request.headers.get("Authorization").split(" ")[-1].encode('utf-8')).decode(
            'utf-8').split(":")
        username = bleach.clean(username)
        user = app.security.datastore.find_user(username=username)

        try:
            if uid:
                eud = EUD()
                eud.uid = uid
                eud.user_id = user.id

                db.session.add(eud)
                db.session.commit()
        except sqlalchemy.exc.IntegrityError:
            db.session.rollback()
            eud = db.session.execute(db.session.query(EUD).filter_by(uid=uid)).first()[0]
            if user and not eud.user_id:
                eud.user_id = user.id
                db.session.add(eud)
                db.session.commit()

        try:
            certificate = Certificate()
            if uid:
                certificate.common_name = common_name
                certificate.eud_uid = uid
                certificate.callsign = eud.callsign
                certificate.expiration_date = datetime.datetime.today() + datetime.timedelta(
                    days=app.config.get("OTS_CA_EXPIRATION_TIME"))
                certificate.server_address = urlparse(request.url_root).hostname
                certificate.server_port = app.config.get("OTS_MARTI_HTTPS_PORT")
                certificate.truststore_filename = os.path.join(app.config.get("OTS_CA_FOLDER"), "truststore-root.p12")
                certificate.user_cert_filename = os.path.join(app.config.get("OTS_CA_FOLDER"), "certs", common_name,
                                                              common_name + ".pem")
                certificate.csr = os.path.join(app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name + ".csr")
                certificate.cert_password = app.config.get("OTS_CA_PASSWORD")

                db.session.add(certificate)
                db.session.commit()
        except sqlalchemy.exc.IntegrityError:
            db.session.rollback()
            if uid:
                certificate = db.session.execute(db.session.query(Certificate).filter_by(eud_uid=eud.uid)).scalar_one()
                certificate.common_name = common_name
                certificate.callsign = eud.callsign
                certificate.expiration_date = datetime.datetime.today() + datetime.timedelta(
                    days=app.config.get("OTS_CA_EXPIRATION_TIME"))
                certificate.server_address = urlparse(request.url_root).hostname
                certificate.server_port = app.config.get("OTS_MARTI_HTTPS_PORT")
                certificate.truststore_filename = os.path.join(app.config.get("OTS_CA_FOLDER"), "truststore-root.p12")
                certificate.user_cert_filename = os.path.join(app.config.get("OTS_CA_FOLDER"), "certs", common_name,
                                                              common_name + ".pem")
                certificate.csr = os.path.join(app.config.get("OTS_CA_FOLDER"), "certs", common_name, common_name + ".csr")
                certificate.cert_password = app.config.get("OTS_CA_PASSWORD")

                db.session.commit()

        if request.headers.get('Accept') == 'text/plain':
            return response, 200, {'Content-Type': 'text/plain', 'Content-Encoding': 'charset=UTF-8'}
        elif request.headers.get('Accept') == 'application/json' or request.headers.get('Accept') == "*/*":
            return jsonify(response)
        else:
            return response, 200, {'Content-Type': 'application/xml', 'Content-Encoding': 'charset=UTF-8'}
    except BaseException as e:
        logger.error(traceback.format_exc())
        return jsonify({'error': str(e)}), 500



from __future__ import annotations

from abc import ABC
from abc import abstractmethod


class BasePlugin(ABC):
    """
    Base class for all plugin types

    The `activate()` method must be implemented and receives the Flask instance, logger, and DB.
    """

    PLUGIN_API_VERSION = "1.0.0"

    @property
    @abstractmethod
    def group(self) -> str:
        """
        Name of entrypoint group the plugin belongs to.
        """


from threading import Thread

from flask import Flask
from pika.channel import Channel

import flask_sqlalchemy
import pika

from opentakserver.extensions import *


class RabbitMQClient:
    def __init__(self, context: Flask):
        self.context = context
        self.logger = logger
        self.db: flask_sqlalchemy.SQLAlchemy = db
        self.socketio = socketio

        self.online_euds = {}
        self.online_callsigns = {}
        self.exchanges = []

        try:
            rabbit_credentials = pika.PlainCredentials(self.context.app.config.get("OTS_RABBITMQ_USERNAME"), self.context.app.config.get("OTS_RABBITMQ_PASSWORD"))
            rabbit_host = self.context.app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
            self.rabbit_connection = pika.SelectConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials),
                                                           self.on_connection_open)
            self.rabbit_channel: Channel = None
            self.iothread = Thread(target=self.rabbit_connection.ioloop.start)
            self.iothread.daemon = True
            self.iothread.start()
            self.is_consuming = False
        except BaseException as e:
            self.logger.error("Failed to connect to rabbitmq: {}".format(e))
            return

    def on_connection_open(self, connection):
        self.rabbit_connection.channel(on_open_callback=self.on_channel_open)
        self.rabbit_connection.add_on_close_callback(self.on_close)

    def on_channel_open(self, channel):
        raise NotImplemented

    def on_close(self, channel, error):
        self.logger.error("cot_controller closing RabbitMQ connection: {}".format(error))

    def on_message(self, unused_channel, basic_deliver, properties, body):
        raise NotImplemented



from flask import Blueprint, request, jsonify
from opentakserver.extensions import db, logger
from opentakserver.models.EUD import EUD

contacts_api = Blueprint("contacts_api", __name__)


@contacts_api.route('/Marti/api/contacts/all')
def get_all_contacts():
    euds = db.session.execute(db.session.query(EUD)).all()

    response = []

    for eud in euds:
        eud = eud[0]
        team_name = eud.team.name if eud.team else 'Cyan'
        team_role = eud.team_role or 'Team Member'
        username = eud.user.username if eud.user else ''
        response.append(
            {'filterGroups': [], 'notes': username, 'callsign': eud.callsign, 'team': team_name,
             'role': team_role, 'takv': f"{eud.platform} {eud.version}", 'uid': eud.uid}
        )

    return jsonify(response)



import traceback

import bleach
import pika
import sqlalchemy.exc
from flask import Blueprint, request, jsonify, current_app as app, Response
from flask_babel import gettext
from flask_login import current_user
from flask_security import roles_required, auth_required

from opentakserver.extensions import db, logger, ldap_manager
from opentakserver.blueprints.ots_api.api import search, paginate
from opentakserver.models.Group import Group
from opentakserver.models.GroupUser import GroupUser

group_api = Blueprint("group_api", __name__)


@group_api.route('/api/groups')
@roles_required("administrator")
def get_groups():
    """ Search groups with filters and pagination

    :parameter: name
    :parameter: type
    :parameter: bitpos
    :parameter: active
    :parameter: page
    :parameter: per_page

    :return: JSON array of groups
    """

    if app.config.get("OTS_ENABLE_LDAP"):
        return jsonify(
            {"success": False, "error": gettext(u"LDAP is enabled. Please view and edit groups on your LDAP server")}), 400

    query = db.session.query(Group)
    query = search(query, Group, 'name')
    query = search(query, Group, 'type')
    query = search(query, Group, 'bitpos')
    query = search(query, Group, 'active')

    return paginate(query)


@group_api.route('/api/groups/all', methods=["GET"])
@auth_required()
def get_all_groups():
    """ Get a list of all groups

    :return: JSON array of groups
    :rtype: Response
    """
    return_value = []

    if app.config.get("OTS_ENABLE_LDAP"):

        groups = ldap_manager.get_user_groups(current_user.username)
        for group in groups:
            if group['cn'].lower().startswith(app.config.get("OTS_LDAP_GROUP_PREFIX").lower()) and not \
                    (group['cn'].lower().endswith("_read") or group['cn'].lower().endswith("_write")):

                g = Group()
                g.id = group['entryuuid']
                g.name = group['cn']
                g.distinguishedName = group['dn']
                g.type = Group.LDAP

                return_value.append(g.to_json())

        return jsonify(return_value)

    if not current_user.has_role("administrator"):
        groups = db.session.execute(
            db.session.query(GroupUser).filter_by(user_id=current_user.id, direction=Group.OUT)).scalars()
        # Make sure a group is only added once, not twice for both IN and OUT
        group_names = []
        for group in groups:
            if group.group.name not in group_names:
                group_names.append(group.group.name)
            else:
                continue
            return_value.append(group.group.to_json())

    else:
        groups = db.session.execute(db.session.query(Group)).scalars()
        for group in groups:
            return_value.append(group.to_json())

    return jsonify(return_value)


@group_api.route('/api/groups/members')
@roles_required("administrator")
def get_group_members():
    """ Get a list of members of a group

    :parameter: name

    :return: JSON array of group members
    :rtype: Response
    """
    if app.config.get("OTS_ENABLE_LDAP"):
        return jsonify(
            {"success": False, "error": gettext(u"LDAP is enabled. Please view and edit groups on your LDAP server")}), 400

    group_name = request.args.get("name")
    if not group_name:
        return jsonify({"success": False, "error": "Please specify a group name"}), 400

    group_name = bleach.clean(group_name)
    group = db.session.execute(db.session.query(Group).filter_by(name=group_name)).first()
    if not group:
        return jsonify({"success": False, "error": f"Group {group_name} not found"}), 404

    group = group[0]
    members = db.session.execute(db.session.query(GroupUser).filter_by(group_id=group.id)).all()
    return_value = []
    for member in members:
        member = member[0]
        return_value.append({"username": member.user.username, "direction": member.direction, "active": member.enabled})

    return return_value


@group_api.route('/api/groups/members', methods=["DELETE"])
@roles_required("administrator")
def remove_user_from_group():
    if app.config.get("OTS_ENABLE_LDAP"):
        return jsonify(
            {"success": False, "error": gettext(u"LDAP is enabled. Please view and edit groups on your LDAP server")}), 400

    username = request.args.get("username")
    group_name = request.args.get("group_name")
    direction = request.args.get("direction")

    if not username or not group_name or not direction:
        return jsonify({"success": False, "error": gettext(u"Please provide the username, group name, and direction")}), 400

    username = bleach.clean(username)
    group_name = bleach.clean(group_name)
    direction = bleach.clean(direction)

    if direction != Group.IN and direction != Group.OUT:
        return jsonify({"success": False, "error": gettext(u"Invalid direction: %(direction)s", direction=direction)}), 400

    user = app.security.datastore.find_user(username=username)
    if not user:
        return jsonify({"success": False, "error": gettext(u"User %(username)s not found", username=username)}), 404

    group = db.session.execute(db.session.query(Group).filter_by(name=group_name)).first()
    if not group:
        return jsonify({"success": False, "error": gettext(u"Group %(group_name)s not found")}), 404

    try:
        GroupUser.query.filter_by(group_id=group[0].id, user_id=user.id, direction=direction).delete()
        db.session.commit()

        rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"),
                                                   app.config.get("OTS_RABBITMQ_PASSWORD"))
        rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
        rabbit_connection = pika.BlockingConnection(
            pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
        channel = rabbit_connection.channel()
        for eud in user.euds:
            channel.queue_unbind(exchange="groups", queue=eud.uid, routing_key=f"{group_name}.{direction}")

        channel.close()
        rabbit_connection.close()

        return jsonify({"success": True})
    except BaseException as e:
        logger.error(f"Failed to remove {username} from {group_name}: {e}")
        logger.debug(traceback.format_exc())
        return jsonify({"success": False, "error": gettext(u"Failed to remove %(username)s from %(group_name)s: %(e)s",
                                                           username=username, group_name=group_name, e=str(e))}), 500


@group_api.route('/api/groups', methods=["POST"])
@roles_required("administrator")
def add_group():
    """ Creates a new group

    :return: 400 if LDAP is enabled, the request is missing the name key or the group exists. 500 on server errors.
    :rtype: Response
    """
    if app.config.get("OTS_ENABLE_LDAP"):
        return jsonify({'success': False, 'error': gettext(u'LDAP is enabled, please use your LDAP server to add groups')}), 400

    if "name" not in request.json.keys():
        return jsonify({'success': False, 'error': gettext(u'Missing name')}), 400

    name = bleach.clean(request.json.get("name"))
    description = bleach.clean(request.json.get("description")) if "description" in request.json.keys() else None

    group = db.session.execute(db.session.query(Group).filter_by(name=name)).first()

    try:
        if not group:
            group = Group()
            group.name = name
            group.type = Group.SYSTEM
            group.description = description
            db.session.add(group)
            db.session.commit()
        else:
            return jsonify({'success': False, 'error': gettext("%(name)s group already exists", name=name)}), 400

    except BaseException as e:
        logger.error(f"Failed to add {name} group: {e}")
        logger.debug(traceback.format_exc())
        return jsonify({'success': False, "error": gettext(u"Failed to add %(name)s group: %(e)s", name=name, e=str(e))}), 500

    return jsonify({'success': True})


@group_api.route('/api/groups', methods=["PUT"])
@roles_required("administrator")
def add_user_to_group():
    """ Adds a users to a group. This will allow all the user's EUDs to subscribe and unsubscribe from the channels/groups they're allowed to see.
    :parameter: users - A list of users to add to a group
    :parameter: group_name - Name of the group to add users to
    :parameter: direction - Group direction, can only be IN or OUT

    :return: 400 if LDAP is enabled, no group or username is specified, or if the specified group or user doesn't exist or the user is already in the group. 200 on success.
    :rtype: Response
    """

    if app.config.get("OTS_ENABLE_LDAP"):
        return jsonify(
            {'success': False, 'error': gettext(u'LDAP is enabled, please use your LDAP server to add users to groups')}), 400

    users = request.json.get("users")
    group_name = request.json.get("group_name")
    direction = request.json.get("direction")

    if users is None or group_name is None or direction is None:
        return jsonify({"success": False, "error": gettext(u"Please provide a list of users, group name, and direction")}), 400

    if direction != "IN" and direction != "OUT":
        return jsonify({"success": False, "error": gettext(u"Direction must be IN or OUT")}), 400

    for username in users:
        user = app.security.datastore.find_user(username=username)
        if not user:
            return jsonify({"success": False, "error": gettext(u"User %(username)s does not exist", username=username)}), 400

        group = db.session.execute(db.session.query(Group).filter_by(name=group_name)).first()
        if not group:
            return jsonify({"success": False, "error": gettext(u"Group %(group_name)s does not exist", group_name=group_name)}), 400

        membership = GroupUser()
        membership.user_id = user.id
        membership.group_id = group[0].id
        membership.direction = direction

        try:
            db.session.add(membership)
            db.session.commit()
        except sqlalchemy.exc.IntegrityError:
            db.session.rollback()

    return jsonify({"success": True})


@group_api.route('/api/groups', methods=["DELETE"])
@roles_required("administrator")
def delete_group():
    if app.config.get("OTS_ENABLE_LDAP"):
        return jsonify(
            {'success': False, 'error': gettext(u'LDAP is enabled, please use your LDAP server to delete groups')}), 400

    if "group_name" not in request.args.keys() or not request.args.get("group_name"):
        return jsonify({'success': False, 'error': gettext(u'Missing group name')}), 400

    group_name = bleach.clean(request.args.get("group_name"))
    if group_name == "__ANON__":
        return jsonify({'success': False, 'error': gettext(u'The __ANON__ group cannot be deleted')}), 400

    try:
        group = db.session.execute(
            db.session.query(Group).filter_by(name=group_name)).first()
        if not group:
            return jsonify({"success": False, "error": gettext(u"No such group: %(group_name)s", group_name=request.args.get('group_name'))}), 404

        group = group[0]

        GroupUser.query.filter_by(group_id=group.id).delete()
        db.session.delete(group)
        db.session.commit()
    except BaseException as e:
        logger.error(f"Failed to delete {request.args.get('group_name')}: {e}")
        logger.debug(traceback.format_exc())
        return jsonify({'success': False, 'error': gettext(u"Failed to delete %(group_name)s: %(e)s",
                                                           group_name=request.args.get('group_name'), e=str(e))}), 500

    return jsonify({'success': True})



"""Meshtastic Support

Revision ID: d91957bb59a0
Revises: 4c7909c34d4e
Create Date: 2024-05-24 04:16:20.765489

"""
from alembic import op
import sqlalchemy as sa
import logging
logger = logging.getLogger('OpenTAKServer')
logger.disabled = False


# revision identifiers, used by Alembic.
revision = 'd91957bb59a0'
down_revision = '4c7909c34d4e'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    try:
        op.create_table('iconsets',
        sa.Column('id', sa.INTEGER(), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('uid', sa.String(length=255), nullable=True),
        sa.Column('version', sa.Integer(), nullable=True),
        sa.Column('defaultFriendly', sa.String(length=255), nullable=True),
        sa.Column('defaultHostile', sa.String(length=255), nullable=True),
        sa.Column('defaultNeutral', sa.String(length=255), nullable=True),
        sa.Column('defaultUnknown', sa.String(length=255), nullable=True),
        sa.Column('selectedGroup', sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint('id')
        )
    except sa.exc.OperationalError:
        logger.warning('iconsets table already exists, skipping')

    try:
        op.create_table('meshtastic_channels',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('psk', sa.String(length=255), nullable=True),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('uplink_enabled', sa.Boolean(), nullable=False),
        sa.Column('downlink_enabled', sa.Boolean(), nullable=False),
        sa.Column('position_precision', sa.Integer(), nullable=False),
        sa.Column('lora_region', sa.Integer(), nullable=False),
        sa.Column('lora_hop_limit', sa.Integer(), nullable=False),
        sa.Column('lora_tx_enabled', sa.Boolean(), nullable=False),
        sa.Column('lora_tx_power', sa.Integer(), nullable=False),
        sa.Column('lora_sx126x_rx_boosted_gain', sa.Boolean(), nullable=False),
        sa.Column('modem_preset', sa.Integer(), nullable=False),
        sa.Column('url', sa.String(length=255), nullable=False),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('url')
        )

        with op.batch_alter_table('euds', schema=None) as batch_op:
            batch_op.add_column(sa.Column('meshtastic_id', sa.Integer(), nullable=True))
            batch_op.add_column(sa.Column('meshtastic_macaddr', sa.String(length=255), nullable=True))
    except sa.exc.OperationalError:
        logger.warning('Meshtastic table already exists, skipping')

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    with op.batch_alter_table('euds', schema=None) as batch_op:
        batch_op.drop_column('meshtastic_macaddr')
        batch_op.drop_column('meshtastic_id')

    op.drop_table('meshtastic_channels')
    op.drop_table('iconsets')
    # ### end Alembic commands ###
